from anoncreds.protocol.globals import LARGE_VPRIME_PRIME, LARGE_E_START, \
    LARGE_E_END_RANGE, LARGE_PRIME
from anoncreds.protocol.types import PublicKey, SecretKey, PrimaryClaim, ID, \
    Attribs, ClaimAttributeValues
from anoncreds.protocol.utils import get_prime_in_range, strToCryptoInteger, \
    randomQR
from anoncreds.protocol.wallet.issuer_wallet import IssuerWallet
from config.config import cmod
from typing import Dict


class PrimaryClaimIssuer:
    def __init__(self, wallet: IssuerWallet):
        self._wallet = wallet

    async def genKeys(self, schemaId: ID, p_prime=None, q_prime=None) -> (
            PublicKey, SecretKey):
        schema = await self._wallet.getSchema(schemaId)
        if not schema.attrNames and isinstance(schema.attrNames, list):
            raise ValueError("List of attribute names is required to "
                             "setup credential definition")

        p_prime = p_prime if p_prime else PrimaryClaimIssuer._genPrime()
        p = 2 * p_prime + 1

        q_prime = q_prime if q_prime else PrimaryClaimIssuer._genPrime()
        q = 2 * q_prime + 1

        n = p * q

        # Generate a random quadratic number
        S = randomQR(n)

        # Generate random numbers corresponding to every attributes
        Xz = PrimaryClaimIssuer._genX(p_prime, q_prime)
        Xr = {}

        for name in schema.attrNames:
            Xr[str(name)] = PrimaryClaimIssuer._genX(p_prime, q_prime)

        # Generate `Z` as the exponentiation of the quadratic random 'S' .
        # over the random `Xz` in the group defined by modulus `n`
        Z = (S ** Xz) % n

        # Generate random numbers corresponding to every attributes
        R = {}
        for name in schema.attrNames:
            R[str(name)] = (S ** Xr[str(name)]) % n

        # Rms is a random number needed corresponding to master secret m1
        Rms = (S ** PrimaryClaimIssuer._genX(p_prime, q_prime)) % n

        # Rctxt is a random number needed corresponding to context attribute m2
        Rctxt = (S ** PrimaryClaimIssuer._genX(p_prime, q_prime)) % n

        return PublicKey(n, Rms, Rctxt, R, S, Z), SecretKey(p_prime, q_prime)

    @classmethod
    def _genX(cls, p_prime, q_prime):
        maxValue = p_prime * q_prime - 1
        minValue = 2
        return cmod.integer(cmod.random(maxValue - minValue)) + minValue

    @classmethod
    def _genPrime(cls):
        # Generate 2 large primes `p_prime` and `q_prime` and use them
        # to generate another 2 primes `p` and `q` of 1024 bits
        prime = cmod.randomPrime(LARGE_PRIME)
        i = 0
        while not cmod.isPrime(2 * prime + 1):
            prime = cmod.randomPrime(LARGE_PRIME)
            i += 1
        print("In {} iterations, found prime {}".format(i, prime))
        return prime

    async def issuePrimaryClaim(self, schemaId: ID, attributes: Attribs,
                                U) -> (PrimaryClaim, Dict[str, ClaimAttributeValues]):
        u = strToCryptoInteger(U) if isinstance(U, str) else U

        if not u:
            raise ValueError("u must be provided to issue a credential")
        # Generate a random prime and
        # Set the Most-significant-bit to 1
        vprimeprime = cmod.integer(cmod.randomBits(LARGE_VPRIME_PRIME) |
                                   (2 ** (LARGE_VPRIME_PRIME - 1)))
        # Generate prime number in the range (2^596, 2^596 + 2^119)
        estart = 2 ** LARGE_E_START
        eend = (estart + 2 ** LARGE_E_END_RANGE)
        e = get_prime_in_range(estart, eend)
        encodedAttrs = attributes.encoded()
        A = await self._sign(schemaId, encodedAttrs, vprimeprime, u, e)

        m2 = await self._wallet.getContextAttr(schemaId)
        claimAttributes = \
            {attr: ClaimAttributeValues(
                attributes._vals[attr], encodedAttrs[attr]) for attr in attributes.keys()}

        return (PrimaryClaim(m2, A, e, vprimeprime), claimAttributes)

    async def _sign(self, schemaId: ID, attrs, v, u, e):
        pk = await self._wallet.getPublicKey(schemaId)
        sk = await self._wallet.getSecretKey(schemaId)
        m2 = await self._wallet.getContextAttr(schemaId)

        Rx = 1 % pk.N
        # Get the product sequence for the (R[i] and attrs[i]) combination
        for k, val in attrs.items():
            Rx = Rx * (pk.R[str(k)] ** val)
        Rx = Rx * (pk.Rctxt ** m2)
        if u != 0:
            u = u % pk.N
            Rx *= u
        nprime = sk.pPrime * sk.qPrime
        einverse = e % nprime
        Q = pk.Z / (Rx * (pk.S ** v)) % pk.N
        A = Q ** (einverse ** -1) % pk.N
        return A

    def __repr__(self):
        return str(self.__dict__)



import pytest

from anoncreds.protocol.types import PublicKey, Schema, Claims, \
    ProofRequest, PredicateGE, FullProof, \
    SchemaKey, ClaimRequest, Proof, AttributeInfo, ProofInfo, AggregatedProof, RequestedProof, PrimaryProof, \
    PrimaryEqualProof, PrimaryPredicateGEProof, ID, ClaimAttributeValues
from config.config import cmod


def testSchemaKeyFromToDict():
    schemaKey = SchemaKey(name='schemaName', version='1.0',
                          issuerId='issuer1')
    assert schemaKey == SchemaKey.fromStrDict(
        schemaKey.toStrDict())


def testSchemaFromToDict():
    schema = Schema(name='schemaName',
                    version='1.0',
                    attrNames=['attr1', 'attr2', 'attr3'],
                    issuerId='issuer1')
    assert schema == Schema.fromStrDict(schema.toStrDict())


def testPKFromToDict():
    pk = PublicKey(N=cmod.integer(11),
                   Rms=cmod.integer(12),
                   Rctxt=cmod.integer(13),
                   R={'a': cmod.integer(1), 'b': cmod.integer(2)},
                   S=cmod.integer(14),
                   Z=cmod.integer(15))

    assert pk == PublicKey.fromStrDict(pk.toStrDict())


def test_pk_from_to_dict():
    pk = PublicKey(N=cmod.integer(12345),
                   Rms=cmod.integer(12) % cmod.integer(12345),
                   Rctxt=cmod.integer(13) % cmod.integer(12345),
                   R={'name': cmod.integer(1) % cmod.integer(
                       12345), 'age': cmod.integer(2) % cmod.integer(12345)},
                   S=cmod.integer(14) % cmod.integer(12345),
                   Z=cmod.integer(15) % cmod.integer(12345))

    pk_serialized = {
        'n': '12345',
        'rms': '12',
        'rctxt': '13',
        'r': {
            'name': '1',
            'age': '2'
        },
        's': '14',
        'z': '15',
    }

    assert pk.to_str_dict() == pk_serialized
    assert pk == PublicKey.from_str_dict(pk_serialized)


def test_claim_request_from_to_dict():
    n = cmod.integer(12345)
    u = cmod.integer(12) % n
    prover_did = '123456789'
    claim_request = ClaimRequest(userId=prover_did, U=u, Ur=None)

    claim_request_serialized = {
        'prover_did': '123456789',
        'u': '12',
        'ur': None
    }

    assert claim_request.to_str_dict() == claim_request_serialized
    assert claim_request == ClaimRequest.from_str_dict(
        claim_request_serialized, n)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
def testRequestClaimsFromToDict(claimsRequestProver1Gvt):
    assert claimsRequestProver1Gvt == ClaimRequest.fromStrDict(
        claimsRequestProver1Gvt.toStrDict())


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
def testClaimsFromToDict(claimsProver1Gvt):
    assert claimsProver1Gvt == Claims.fromStrDict(claimsProver1Gvt.toStrDict())


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
def testClaimsFromToDictPrimaryOnly(claimsProver1Gvt):
    claims = Claims(primaryClaim=claimsProver1Gvt.primaryClaim)
    assert claims == Claims.fromStrDict(claims.toStrDict())


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testAggregatedProofFromToDict(prover1, nonce, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})

    proof = await prover1.presentProof(proofRequest)

    assert proof.aggregatedProof == AggregatedProof.fromStrDict(
        proof.aggregatedProof.toStrDict())


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testRequestedProofFromToDict(prover1, nonce, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})

    proof = await prover1.presentProof(proofRequest)

    assert proof.requestedProof == RequestedProof.fromStrDict(
        proof.requestedProof.toStrDict())


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testClaimProofFromToDict(prover1, nonce, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})

    proof = await prover1.presentProof(proofRequest)

    assert proof == FullProof.fromStrDict(proof.toStrDict())


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testClaimProofFromToDictPrimaryOnly(prover1, nonce, claimsProver1Gvt, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})

    proof = await prover1.presentProof(proofRequest)

    proofInfo = proof.proofs[str(schemaGvt.seqId)]
    proofs = {schemaGvt.seqId: ProofInfo(Proof(primaryProof=proofInfo.proof.primaryProof),
                                         issuer_did=schemaGvt.issuerId,
                                         schema_seq_no=proofInfo.schema_seq_no)}
    proof = proof._replace(proofs=proofs)
    assert proof == FullProof.fromStrDict(proof.toStrDict())


def test_proof_request_from_to_dict():
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})

    proof_input_serialized = {
        'name': 'proof1',
        'version': '1.0',
        'nonce': '1',
        'requested_attrs': {'attr_uuid': {'name': 'name', 'schema_seq_no': None, 'issuer_did': None}},
        'requested_predicates': {'predicate_uuid': {'p_type': 'GE', 'value': 18, 'attr_name': 'age', 'schema_seq_no': None,
                                                    'issuer_did': None}}
    }
    assert proofRequest.to_str_dict() == proof_input_serialized
    assert proofRequest == ProofRequest.from_str_dict(proof_input_serialized)
    assert proofRequest == ProofRequest.from_str_dict(
        proofRequest.to_str_dict())


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def test_requested_proof_from_to_dict(prover1, nonce, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})

    proof = await prover1.presentProof(proofRequest)

    requested_proof_serialized = {
        'revealed_attrs': {'attr_uuid': ['1', 'Alex', '1139481716457488690172217916278103335']},
        'predicates': {'predicate_uuid': '1'},
        'self_attested_attrs': {},
        'unrevealed_attrs': {}
    }

    assert proof.requestedProof.to_str_dict() == requested_proof_serialized
    assert proof.requestedProof == RequestedProof.from_str_dict(
        requested_proof_serialized)
    assert proof.requestedProof == RequestedProof.from_str_dict(
        proof.requestedProof.to_str_dict())


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def test_attribute_values_from_to_dict():
    attr_values = ClaimAttributeValues(raw='Alex', encoded=cmod.integer(11))

    attr_values_serialized = ['Alex', '11']

    assert attr_values.to_str_dict() == attr_values_serialized
    assert attr_values == ClaimAttributeValues.from_str_dict(
        attr_values_serialized)
    assert attr_values == ClaimAttributeValues.from_str_dict(
        attr_values.to_str_dict())


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def test_aggregated_proof_from_to_dict(prover1, nonce, claimsProver1Gvt):
    aggregated_proof = AggregatedProof(
        1, [cmod.integer(111), cmod.integer(32321), cmod.integer(323)])

    aggregated_proof_serialized = {
        'c_hash': '1',
        'c_list': [[111], [126, 65], [1, 67]]
    }

    assert aggregated_proof.to_str_dict() == aggregated_proof_serialized
    assert aggregated_proof == AggregatedProof.from_str_dict(
        aggregated_proof_serialized)
    assert aggregated_proof == AggregatedProof.from_str_dict(
        aggregated_proof.to_str_dict())


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def test_equal_proof_from_to_dict():
    n = cmod.integer(12345)

    eqProof = PrimaryEqualProof(e=cmod.integer(1), v=cmod.integer(11), m={'name': cmod.integer(12)},
                                m1=cmod.integer(12), m2=cmod.integer(32), Aprime=cmod.integer(32) % n,
                                revealedAttrs={'name': cmod.integer(35)})

    proof_serialized = {
        'a_prime': '32',
        'e': '1',
        'm': {'name': '12'},
        'm1': '12',
        'm2': '32',
        'v': '11',
        'revealed_attrs': {'name': '35'}
    }

    assert eqProof.to_str_dict() == proof_serialized
    assert eqProof == PrimaryEqualProof.from_str_dict(proof_serialized, n)
    assert eqProof == PrimaryEqualProof.from_str_dict(eqProof.to_str_dict(), n)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def test_ge_proof_from_to_dict():
    n = cmod.integer(12345)

    predicate = PredicateGE(attrName='age', value=18)
    geProof = PrimaryPredicateGEProof(alpha=cmod.integer(1), mj=cmod.integer(12), r={'1': cmod.integer(13)},
                                      u={'1': cmod.integer(42)}, T={'1': cmod.integer(21) % n}, predicate=predicate)

    proof_serialized = {
        'alpha': '1',
        'mj': '12',
        't': {'1': '21'},
        'r': {'1': '13'},
        'u': {'1': '42'},
        'predicate': {
            'p_type': 'GE',
            'attr_name': 'age',
            'value': 18,
            'schema_seq_no': None,
            'issuer_did': None
        }
    }

    assert geProof.to_str_dict() == proof_serialized
    assert geProof == PrimaryPredicateGEProof.from_str_dict(
        proof_serialized, n)
    assert geProof == PrimaryPredicateGEProof.from_str_dict(
        geProof.to_str_dict(), n)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def test_primary_proof_from_to_dict():
    n = cmod.integer(12345)

    eqProof = PrimaryEqualProof(e=cmod.integer(1), v=cmod.integer(11), m={'name': cmod.integer(12)},
                                m1=cmod.integer(12), m2=cmod.integer(32), Aprime=cmod.integer(32) % n,
                                revealedAttrs={'name': cmod.integer(35)})

    predicate = PredicateGE(attrName='age', value=18)
    geProof = PrimaryPredicateGEProof(alpha=cmod.integer(1), mj=cmod.integer(12), r={'1': cmod.integer(13)},
                                      u={'1': cmod.integer(42)}, T={'1': cmod.integer(21) % n}, predicate=predicate)
    primaryProof = PrimaryProof(eqProof=eqProof, geProofs=[geProof])

    proof_serialized = {
        'eq_proof': {
            'a_prime': '32',
            'e': '1',
            'm': {'name': '12'},
            'm1': '12',
            'm2': '32',
            'v': '11',
            'revealed_attrs': {'name': '35'}
        },
        'ge_proofs': [
            {
                'alpha': '1',
                'mj': '12',
                't': {'1': '21'},
                'r': {'1': '13'},
                'u': {'1': '42'},
                'predicate': {
                    'p_type': 'GE',
                    'attr_name': 'age',
                    'value': 18,
                    'schema_seq_no': None,
                    'issuer_did': None
                }
            }
        ]
    }

    assert primaryProof.to_str_dict() == proof_serialized
    assert primaryProof == PrimaryProof.from_str_dict(proof_serialized, n)
    assert primaryProof == PrimaryProof.from_str_dict(
        primaryProof.to_str_dict(), n)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def test_proof_info_from_to_dict():
    n = cmod.integer(12345)

    eqProof = PrimaryEqualProof(e=cmod.integer(1), v=cmod.integer(11), m={'name': cmod.integer(12)},
                                m1=cmod.integer(12), m2=cmod.integer(32), Aprime=cmod.integer(32) % n,
                                revealedAttrs={'name': cmod.integer(35)})

    predicate = PredicateGE(attrName='age', value=18)
    geProof = PrimaryPredicateGEProof(alpha=cmod.integer(1), mj=cmod.integer(12), r={'1': cmod.integer(13)},
                                      u={'1': cmod.integer(42)}, T={'1': cmod.integer(21) % n}, predicate=predicate)
    primaryProof = PrimaryProof(eqProof=eqProof, geProofs=[geProof])
    proofInfo = Proof(primaryProof=primaryProof)
    proof = ProofInfo(schema_seq_no=1, proof=proofInfo, issuer_did='did')

    proof_serialized = {
        'issuer_did': 'did',
        'schema_seq_no': 1,
        'proof': {
            'primary_proof': {
                'eq_proof': {
                    'a_prime': '32',
                    'e': '1',
                    'm': {'name': '12'},
                    'm1': '12',
                    'm2': '32',
                    'v': '11',
                    'revealed_attrs': {'name': '35'}
                },
                'ge_proofs': [
                    {
                        'alpha': '1',
                        'mj': '12',
                        't': {'1': '21'},
                        'r': {'1': '13'},
                        'u': {'1': '42'},
                        'predicate': {
                            'p_type': 'GE',
                            'attr_name': 'age',
                            'value': 18,
                            'schema_seq_no': None,
                            'issuer_did': None
                        }
                    }
                ]
            }
        }
    }

    assert proof.to_str_dict() == proof_serialized
    assert proof == ProofInfo.from_str_dict(proof_serialized, n)
    assert proof == ProofInfo.from_str_dict(proof.to_str_dict(), n)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def test_proof_from_to_dict(prover1, nonce, claimsProver1Gvt, schemaGvt):
    n = (await prover1.wallet.getPublicKey(ID(schemaId=schemaGvt.seqId))).N
    proofRequest = ProofRequest("proof1", "1.0", nonce,
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})

    proof = await prover1.presentProof(proofRequest)

    proofInfo = proof.proofs[str(schemaGvt.seqId)]
    proof = ProofInfo(Proof(primaryProof=proofInfo.proof.primaryProof),
                      issuer_did=schemaGvt.issuerId,
                      schema_seq_no=proofInfo.schema_seq_no)

    assert proof == ProofInfo.from_str_dict(proof.to_str_dict(), n)



"""
Package metadata
"""
__version_info__ = (1, 0, 31)
__version__ = '.'.join(map(str, __version_info__))
__author__ = "Hyperledger"
__license__ = "Apache 2.0"

__all__ = ['__version_info__', '__version__', '__author__', '__license__']



from anoncreds.protocol.globals import LARGE_PUBLIC_RHO, LARGE_PUBLIC_B
from anoncreds.protocol.types import PublicParams
from config.config import cmod


class PublicParamsBuilder:
    def __init__(self):
        super().__init__()

    @classmethod
    def _genRhoBGamma(cls):
        while True:
            rho = cmod.randomPrime(LARGE_PUBLIC_RHO)
            b = cmod.randomBits(LARGE_PUBLIC_B)
            Gamma = b * rho + 1
            if cmod.isPrime(Gamma) and (rho % b != 0):
                return rho, b, Gamma

    @classmethod
    def _genG(cls, Gamma, b):
        while True:
            gprime = cmod.random(Gamma)
            g = (gprime ** b) % Gamma
            if g != 1:
                return g

    @classmethod
    def generateParams(cls) -> PublicParams:
        rho, b, Gamma = PublicParamsBuilder._genRhoBGamma()
        g = PublicParamsBuilder._genG(Gamma, b)
        r = cmod.random(rho)
        h = g ** r
        return PublicParams(Gamma, rho, g, h)



import pytest

from anoncreds.protocol.types import ProofRequest, ProofClaims, PredicateGE, RequestedProof, AttributeInfo
from anoncreds.protocol.utils import encodeAttr


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testEmpty(prover1):
    proofRequest = ProofRequest(
        "proof1", "1.0", 1, verifiableAttributes={}, predicates={})
    assert ({}, RequestedProof([], [], [], [])) == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testOneRevealedOnly(prover1, allClaims, schemaGvtId, attrRepo, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1, verifiableAttributes={
                                'uuid': AttributeInfo(name='name')})
    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)

    proofClaims = {schemaGvt.seqId: ProofClaims(claimsGvt, ['name'], [])}
    attr = attrRepo.getAttributes(
        schemaGvtId.schemaKey, prover1.proverId)['name']
    requestedProof = RequestedProof(
        revealed_attrs={'uuid': [str(schemaGvt.seqId), attr, str(encodeAttr(attr))]})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testPredicatesEmpty(prover1, allClaims, schemaGvtId, attrRepo, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={'uuid': AttributeInfo(name='name')}, predicates={})

    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)

    proofClaims = {schemaGvt.seqId: ProofClaims(claimsGvt, ['name'], [])}

    attr = attrRepo.getAttributes(
        schemaGvtId.schemaKey, prover1.proverId)['name']
    requestedProof = RequestedProof(
        revealed_attrs={'uuid': [schemaGvt.seqId, attr, str(encodeAttr(attr))]})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testOnePredicateOnly(prover1, allClaims, schemaGvtId, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1, predicates={
                                'uuid': PredicateGE('age', 18)})

    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)
    proofClaims = {schemaGvt.seqId:
                   ProofClaims(claimsGvt, [], [PredicateGE('age', 18)])}

    requestedProof = RequestedProof(predicates={'uuid': schemaGvt.seqId})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testRevealedEmpty(prover1, allClaims, schemaGvtId, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={}, predicates={'uuid': PredicateGE('age', 18)})

    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)
    proofClaims = {schemaGvt.seqId:
                   ProofClaims(claimsGvt, [], [PredicateGE('age', 18)])}

    requestedProof = RequestedProof(predicates={'uuid': schemaGvt.seqId})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testRevealedAndPredicateSameIssuer(prover1, allClaims, schemaGvtId,
                                             attrRepo, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})

    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)
    proofClaims = {schemaGvt.seqId: ProofClaims(
        claimsGvt, ['name'], [PredicateGE('age', 18)])}

    attr = attrRepo.getAttributes(
        schemaGvtId.schemaKey, prover1.proverId)['name']
    requestedProof = RequestedProof(revealed_attrs={'attr_uuid': [schemaGvt.seqId, attr, str(encodeAttr(attr))]},
                                    predicates={'predicate_uuid': schemaGvt.seqId})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testRevealedAndPredicateDifferentIssuers(prover1, allClaims,
                                                   schemaGvtId, schemaXyzId,
                                                   attrRepo, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='status')},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})

    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)
    claimsXyz = await prover1.wallet.getClaimSignature(schemaXyzId)
    proofClaims = {schemaGvt.seqId: ProofClaims(claimsGvt, [], [PredicateGE('age', 18)]),
                   schemaGvt.seqId: ProofClaims(claimsXyz, ['status'], [])}

    attr = attrRepo.getAttributes(
        schemaXyzId.schemaKey, prover1.proverId)['status']
    requestedProof = RequestedProof(revealed_attrs={'attr_uuid': [schemaGvt.seqId, attr, str(encodeAttr(attr))]},
                                    predicates={'predicate_uuid': schemaGvt.seqId})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testMultipledRevealed(prover1, allClaims, schemaGvtId,
                                schemaXyzId, attrRepo, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={'attr_uuid1': AttributeInfo(name='status'),
                                                      'attr_uuid2': AttributeInfo(name='name')})

    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)
    claimsXyz = await prover1.wallet.getClaimSignature(schemaXyzId)
    proofClaims = {schemaGvt.seqId: ProofClaims(claimsGvt, ['name'], []),
                   schemaGvt.seqId: ProofClaims(claimsXyz, ['status'], [])}

    attr1 = attrRepo.getAttributes(
        schemaXyzId.schemaKey, prover1.proverId)['status']
    attr2 = attrRepo.getAttributes(
        schemaGvtId.schemaKey, prover1.proverId)['name']
    requestedProof = RequestedProof(revealed_attrs={'attr_uuid1': [schemaGvt.seqId, attr1, str(encodeAttr(attr1))],
                                                    'attr_uuid2': [schemaGvt.seqId, attr2, str(encodeAttr(attr2))]})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testMultipledPredicates(prover1, allClaims, schemaGvtId,
                                  schemaXyzId, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('period', 8)})

    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)
    claimsXyz = await prover1.wallet.getClaimSignature(schemaXyzId)
    proofClaims = {schemaGvt.seqId: ProofClaims(claimsGvt, [], [PredicateGE('age', 18)]),
                   schemaGvt.seqId: ProofClaims(claimsXyz, [], [PredicateGE('period', 8)])}

    requestedProof = RequestedProof(predicates={'predicate_uuid1': schemaGvt.seqId,
                                                'predicate_uuid2': schemaGvt.seqId})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testMultipleAll(prover1, allClaims, schemaGvtId, schemaXyzId,
                          attrRepo, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={'attr_uuid1': AttributeInfo(name='status'),
                                                      'attr_uuid2': AttributeInfo(name='name')},
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('period', 8)})

    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)
    claimsXyz = await prover1.wallet.getClaimSignature(schemaXyzId)
    proofClaims = {schemaGvt.seqId: ProofClaims(claimsGvt, ['name'], [PredicateGE('age', 18)]),
                   schemaGvt.seqId: ProofClaims(claimsXyz, ['status'], [PredicateGE('period', 8)])}

    attr1 = attrRepo.getAttributes(
        schemaXyzId.schemaKey, prover1.proverId)['status']
    attr2 = attrRepo.getAttributes(
        schemaGvtId.schemaKey, prover1.proverId)['name']

    requestedProof = RequestedProof(
        revealed_attrs={'attr_uuid1': [schemaGvt.seqId, attr1, str(encodeAttr(attr1))],
                        'attr_uuid2': [schemaGvt.seqId, attr2, str(encodeAttr(attr2))]},
        predicates={'predicate_uuid1': schemaGvt.seqId,
                    'predicate_uuid2': schemaGvt.seqId})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testAttrNotFound(prover1, allClaims):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={'attr_uuid1': AttributeInfo(name='name'),
                                                      'attr_uuid2': AttributeInfo(name='aaa')})
    with pytest.raises(ValueError):
        await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testPredicateNotFound(prover1, allClaims):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('aaaa', 8)})
    with pytest.raises(ValueError):
        await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testOneRevealedFromSchema(prover1, allClaims, schemaGvtId, attrRepo, schemaGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={
                                    'uuid': AttributeInfo(name='name', schema_seq_no=schemaGvt.seqId)})
    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)

    proofClaims = {schemaGvt.seqId: ProofClaims(claimsGvt, ['name'], [])}
    attr = attrRepo.getAttributes(
        schemaGvtId.schemaKey, prover1.proverId)['name']
    requestedProof = RequestedProof(
        revealed_attrs={'uuid': [str(schemaGvt.seqId), attr, str(encodeAttr(attr))]})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testOneRevealedFromOtherSchema(prover1, allClaims, schemaXyz):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={
                                    'uuid': AttributeInfo(name='name', schema_seq_no=schemaXyz.seqId)})

    with pytest.raises(ValueError):
        await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testOneRevealedFromSpecificSchemaAndIssuer(prover1, allClaims, schemaGvt, schemaGvtId, attrRepo, keysGvt):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={'uuid': AttributeInfo(name='name', schema_seq_no=schemaGvt.seqId,
                                                                            issuer_did=schemaGvt.issuerId)})
    claimsGvt = await prover1.wallet.getClaimSignature(schemaGvtId)

    proofClaims = {schemaGvt.seqId: ProofClaims(claimsGvt, ['name'], [])}
    attr = attrRepo.getAttributes(
        schemaGvtId.schemaKey, prover1.proverId)['name']
    requestedProof = RequestedProof(
        revealed_attrs={'uuid': [str(schemaGvt.seqId), attr, str(encodeAttr(attr))]})

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)

    assert proofClaims, requestedProof == await prover1._findClaims(proofRequest)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testOneRevealedFromOtherIssuer(prover1, allClaims, schemaGvt, schemaXyz):
    proofRequest = ProofRequest("proof1", "1.0", 1,
                                verifiableAttributes={'uuid': AttributeInfo(name='name', schema_seq_no=schemaGvt.seqId,
                                                                            issuer_did=schemaXyz.issuerId)})

    with pytest.raises(ValueError):
        await prover1._findClaims(proofRequest)



import pytest

from anoncreds.protocol.types import ProofRequest, ID, AttributeInfo
from anoncreds.protocol.utils import groupIdentityG1
from anoncreds.test.conftest import presentProofAndVerify


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testIssueRevocationCredential(claimsProver1Gvt, issuerGvt,
                                        schemaGvtId):
    nonRevocClaimGvtProver1 = claimsProver1Gvt.nonRevocClaim
    acc = await issuerGvt.wallet.getAccumulator(schemaGvtId)
    tails = await issuerGvt.wallet.getTails(schemaGvtId)
    assert nonRevocClaimGvtProver1
    assert nonRevocClaimGvtProver1.witness
    assert nonRevocClaimGvtProver1.witness.V
    assert nonRevocClaimGvtProver1.i == 1
    assert nonRevocClaimGvtProver1.witness.gi == tails.g[1]

    assert acc.V
    assert acc.acc != 1

    assert nonRevocClaimGvtProver1.witness.V == acc.V


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testRevoce(claimsProver1Gvt, issuerGvt, schemaGvtId):
    await issuerGvt.revoke(schemaGvtId, 1)
    newAcc = await issuerGvt.wallet.getAccumulator(schemaGvtId)
    assert not newAcc.V
    assert newAcc.acc == groupIdentityG1()


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testUpdateWitnessNotChangedIfInSync(claimsProver1Gvt, schemaGvt,
                                              prover1):
    nonRevocClaimGvtProver1 = claimsProver1Gvt.nonRevocClaim
    acc = await prover1.wallet.getAccumulator(ID(schemaId=schemaGvt.seqId))

    # not changed as in sync
    oldOmega = nonRevocClaimGvtProver1.witness.omega

    c2 = await prover1._nonRevocProofBuilder.updateNonRevocationClaim(
        schemaGvt.seqId,
        nonRevocClaimGvtProver1)
    assert c2.witness.V == acc.V
    assert oldOmega == c2.witness.omega


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testUpdateWitnessChangedIfOutOfSync(claimsProver1Gvt, issuerGvt,
                                              schemaGvt, prover1):
    nonRevocClaimGvtProver1 = claimsProver1Gvt.nonRevocClaim
    acc = await issuerGvt.wallet.getAccumulator(ID(schemaId=schemaGvt.seqId))

    # not in sync
    acc.V.add(3)
    assert nonRevocClaimGvtProver1.witness.V != acc.V

    # witness is updated
    oldOmega = nonRevocClaimGvtProver1.witness.omega
    c2 = await prover1._nonRevocProofBuilder.updateNonRevocationClaim(
        schemaGvt.seqId,
        nonRevocClaimGvtProver1)
    assert c2.witness.V == acc.V
    assert oldOmega != c2.witness.omega


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testUpdateRevocedWitness(claimsProver1Gvt, issuerGvt, schemaGvt,
                                   prover1):
    nonRevocClaimGvtProver1 = claimsProver1Gvt.nonRevocClaim
    await issuerGvt.revoke(ID(schemaId=schemaGvt.seqId), 1)
    with pytest.raises(ValueError):
        await prover1._nonRevocProofBuilder.updateNonRevocationClaim(
            schemaGvt.seqId, nonRevocClaimGvtProver1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testInitNonRevocClaim(schemaGvtId, prover1, issuerGvt,
                                attrsProver1Gvt, keysGvt, issueAccumulatorGvt):
    claimsReq = await prover1.createClaimRequest(schemaGvtId)
    claim_signature, claim_attributes = await issuerGvt.issueClaim(schemaGvtId, claimsReq)

    oldV = claim_signature.nonRevocClaim.v
    await prover1.processClaim(schemaGvtId, claim_attributes, claim_signature)
    newC2 = (await prover1.wallet.getClaimSignature(schemaGvtId)).nonRevocClaim
    vrPrime = (
        await prover1.wallet.getNonRevocClaimInitData(schemaGvtId)).vPrime

    assert oldV + vrPrime == newC2.v


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testCAndTauList(claimsProver1Gvt, schemaGvt, prover1):
    nonRevocClaimGvtProver1 = claimsProver1Gvt.nonRevocClaim
    proofRevBuilder = prover1._nonRevocProofBuilder
    assert await proofRevBuilder.testProof(schemaGvt.seqId,
                                           nonRevocClaimGvtProver1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testRevocedWithUpdateWitness(schemaGvtId, issuerGvt, prover1,
                                       verifier, claimsProver1Gvt):
    await issuerGvt.revoke(schemaGvtId, 1)

    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={'attr_uuid': AttributeInfo(name='name')})
    with pytest.raises(ValueError):
        await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testRevocedWithoutUpdateWitness(schemaGvtId, issuerGvt, prover1,
                                          verifier, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={'attr_uuid': AttributeInfo(name='name')})

    proof = await prover1.presentProof(proofRequest)

    await issuerGvt.revoke(schemaGvtId, 1)

    return await verifier.verify(proofRequest, proof)



from anoncreds.protocol.wallet.wallet import Wallet


def test_wallet_name(publicRepo):
    walletName = "test-wallet-name-1"
    wallet = Wallet(walletName, publicRepo)
    assert wallet.name == walletName
    assert wallet.walletId == wallet.name



import socket
import json
import asyncio
import logging

from anoncreds.protocol.issuer import Issuer
from anoncreds.protocol.repo.attributes_repo import AttributeRepoInMemory
from anoncreds.protocol.repo.public_repo import PublicRepoInMemory
from anoncreds.protocol.types import ID, ProofRequest
from anoncreds.protocol.wallet.issuer_wallet import IssuerWalletInMemory
from anoncreds.test.conftest import GVT, primes1
from anoncreds.protocol.prover import Prover
from anoncreds.protocol.wallet.prover_wallet import ProverWalletInMemory

logging.basicConfig(format=u'%(filename)s[LINE:%(lineno)d]# %(levelname)-8s [%(asctime)s]  %(message)s',
                    level=logging.DEBUG)

loop = asyncio.get_event_loop()
global_dict = {}
ip = '127.0.0.1'
port = 1234
chunk_size = 102400


def main():
    sock = socket.socket()
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((ip, port))
    sock.listen(1)
    logging.debug('Listening')
    conn, _ = sock.accept()
    logging.debug('Connected')

    while True:
        data = json.loads(conn.recv(chunk_size).decode("utf-8"))
        logging.debug('received data: {}'.format(data))
        if ('type' in data) & (data['type'] == 'get_claim_def'):
            logging.debug('get_claim_def -> start')
            future = asyncio.ensure_future(init(primes1(), conn))
            loop.run_until_complete(future)
            logging.debug('get_claim_def -> done')
        if ('type' in data) & (data['type'] == 'get_proof'):
            logging.debug('get_proof -> start')
            future = asyncio.ensure_future(create_proof(conn, data['data']))
            loop.run_until_complete(future)
            logging.debug('get_proof -> done')
        if (('type' in data) & (data['type'] == 'close')) | (not data):
            break

    sock.close()


async def init(primes, conn):
    # 1. Init entities
    public_repo = PublicRepoInMemory()
    attr_repo = AttributeRepoInMemory()
    issuer = Issuer(IssuerWalletInMemory('issuer1', public_repo), attr_repo)

    # 2. Create a Schema
    schema = await issuer.genSchema('GVT', '1.0', GVT.attribNames())
    schema_id = ID(schema.getKey())

    # 3. Create keys for the Schema
    await issuer.genKeys(schema_id, **primes)

    # 4. Issue accumulator
    await issuer.issueAccumulator(schemaId=schema_id, iA='110', L=5)

    # 5. set attributes for user1
    prover_id = 'BzfFCYk'
    attributes = GVT.attribs(name='Alex', age=28, height=175, sex='male')
    attr_repo.addAttributes(schema.getKey(), prover_id, attributes)

    # 6. request Claims
    prover = Prover(ProverWalletInMemory(prover_id, public_repo))
    claims_req = await prover.createClaimRequest(schema_id, reqNonRevoc=False)
    (claim_signature, claim_attributes) = await issuer.issueClaim(schema_id, claims_req)

    await prover.processClaim(schema_id, claim_attributes, claim_signature)

    global global_dict
    global_dict = {
        'prover': prover
    }

    public_key = await issuer.wallet.getPublicKey(schema_id)

    conn.send(json.dumps({
        'primary': public_key.to_str_dict(),
        'revocation': None
    }).encode())


async def create_proof(conn, proof_request):
    proof_request = ProofRequest.from_str_dict(proof_request)
    proof = await global_dict['prover'].presentProof(proof_request)

    conn.send(json.dumps(proof.to_str_dict()).encode())

if __name__ == '__main__':
    main()



import socket
import json
import asyncio
import logging

from anoncreds.protocol.issuer import Issuer
from anoncreds.protocol.repo.attributes_repo import AttributeRepoInMemory
from anoncreds.protocol.repo.public_repo import PublicRepoInMemory
from anoncreds.protocol.types import ID, ClaimRequest
from anoncreds.protocol.wallet.issuer_wallet import IssuerWalletInMemory
from anoncreds.test.conftest import GVT, primes1

logging.basicConfig(format=u'%(filename)s[LINE:%(lineno)d]# %(levelname)-8s [%(asctime)s]  %(message)s',
                    level=logging.DEBUG)

loop = asyncio.get_event_loop()
global_dict = {}
ip = '127.0.0.1'
port = 1234
chunk_size = 102400


def main():
    sock = socket.socket()
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((ip, port))
    sock.listen(1)
    logging.debug('Listening')
    conn, _ = sock.accept()
    logging.debug('Connected')

    while True:
        data = json.loads(conn.recv(chunk_size).decode("utf-8"))
        logging.debug('received data: {}'.format(data))
        if ('type' in data) & (data['type'] == 'get_claim_def'):
            logging.debug('get_claim_def -> start')
            init = asyncio.ensure_future(issuer_init(primes1(), conn))
            loop.run_until_complete(init)
            logging.debug('get_claim_def -> done')
        if ('type' in data) & (data['type'] == 'issue_claim'):
            logging.debug('issue_claim -> start')
            future = asyncio.ensure_future(
                issue_claim(conn, data['data']['blinded_ms']))
            loop.run_until_complete(future)
            logging.debug('issue_claim -> done')
        if (('type' in data) & (data['type'] == 'close')) | (not data):
            break

    sock.close()


async def issuer_init(primes, conn):
    # 1. Init entities
    public_repo = PublicRepoInMemory()
    attr_repo = AttributeRepoInMemory()
    issuer = Issuer(IssuerWalletInMemory('issuer1', public_repo), attr_repo)

    # 2. Create a Schema
    schema = await issuer.genSchema('GVT', '1.0', GVT.attribNames())
    schema_id = ID(schema.getKey())

    # 3. Create keys for the Schema
    await issuer.genKeys(schema_id, **primes)

    # 4. Issue accumulator
    await issuer.issueAccumulator(schemaId=schema_id, iA='110', L=5)

    # 4. set attributes for user1
    prover_id = 'BzfFCYk'
    attributes = GVT.attribs(name='Alex', age=28, height=175, sex='male')
    attr_repo.addAttributes(schema.getKey(), prover_id, attributes)

    public_key = await issuer.wallet.getPublicKey(schema_id)

    global global_dict
    global_dict = {
        'schema_id': schema_id,
        'public_key': public_key,
        'issuer': issuer
    }

    conn.send(json.dumps({
        'primary': public_key.to_str_dict(),
        'revocation': None
    }).encode())


async def issue_claim(conn, claim_request):
    claim_request = ClaimRequest.from_str_dict(
        claim_request, global_dict['public_key'].N)

    (signature, claims) = await global_dict['issuer'].issueClaim(global_dict['schema_id'], claim_request)

    msg = {
        'signature': signature.to_str_dict(),
        'claim': {el: claims[el].to_str_dict() for el in claims}
    }

    conn.send(json.dumps(msg).encode())

if __name__ == '__main__':
    main()



from anoncreds.protocol.globals import ITERATIONS, DELTA


def calcTeq(pk, Aprime, e, v, mtilde, m1Tilde, m2Tilde, unrevealedAttrNames):
    Rur = 1 % pk.N
    for k in unrevealedAttrNames:
        Rur = Rur * (pk.R[k] ** mtilde[k])
    Rur *= pk.Rms ** m1Tilde
    Rur *= pk.Rctxt ** m2Tilde
    return ((Aprime ** e) * Rur * (pk.S ** v)) % pk.N


def calcTge(pk, u, r, mj, alpha, T):
    TauList = []
    for i in range(0, ITERATIONS):
        Ttau = (pk.Z ** u[str(i)]) * (pk.S ** r[str(i)]) % pk.N
        TauList.append(Ttau)
    Ttau = (pk.Z ** mj) * (pk.S ** r[DELTA]) % pk.N
    TauList.append(Ttau)

    # gen Q
    Q = 1 % pk.N
    for i in range(0, ITERATIONS):
        Q *= T[str(i)] ** u[str(i)]
    Q = Q * (pk.S ** alpha) % pk.N
    TauList.append(Q)

    return TauList



import pytest

from anoncreds.protocol.types import ProofRequest, PredicateGE, Claims, \
    ProofClaims, AttributeInfo
from anoncreds.test.conftest import presentProofAndVerify


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testPrimaryClaimOnlyEmpty(prover1, verifier, claimsProver1Gvt, nonce):
    proofRequest = ProofRequest("proof1", "1.0", nonce)
    claims, requestedProof = await prover1._findClaims(proofRequest)
    claims = {schemaId: ProofClaims(
        Claims(primaryClaim=proofClaim.claims.primaryClaim))
        for schemaId, proofClaim in claims.items()}

    proof = await prover1._prepareProof(claims, proofRequest.nonce, requestedProof)

    assert await verifier.verify(proofRequest, proof)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testPrimaryClaimNoPredicates(prover1, verifier, claimsProver1Gvt,
                                       nonce, schemaGvtId):
    proofRequest = ProofRequest("proof1", "1.0", nonce,
                                verifiableAttributes={'uuid1': AttributeInfo(name='name')}, predicates={})

    claims, requestedProof = await prover1._findClaims(proofRequest)
    claims = {
        schemaId: ProofClaims(
            Claims(primaryClaim=proofClaim.claims.primaryClaim), [AttributeInfo(name='name')], [])
        for schemaId, proofClaim in claims.items()}
    proof = await prover1._prepareProof(claims, proofRequest.nonce, requestedProof)

    assert await verifier.verify(proofRequest, proof)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testPrimaryClaimPredicatesOnly(prover1, verifier, claimsProver1Gvt,
                                         nonce, schemaGvtId):
    predicate = PredicateGE('age', 18)
    proofRequest = ProofRequest("proof1", "1.0", nonce,
                                verifiableAttributes={},
                                predicates={'predicate_uuid1': predicate})

    claims, requestedProof = await prover1._findClaims(proofRequest)
    claims = {
        schemaId: ProofClaims(
            Claims(primaryClaim=proofClaim.claims.primaryClaim), predicates=[predicate])
        for schemaId, proofClaim in claims.items()}

    proof = await prover1._prepareProof(claims, proofRequest.nonce, requestedProof)

    assert await verifier.verify(proofRequest, proof)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testEmpty(prover1, verifier, claimsProver1Gvt):
    assert await presentProofAndVerify(verifier, ProofRequest("proof1", "1.0", verifier.generateNonce()), prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testNoPredicates(prover1, verifier, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={'uuid': AttributeInfo(name='name')}, predicates={})
    assert await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testMultipleRevealedAttrs(prover1, verifier, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={'uuid1': AttributeInfo(name='name'),
                                                      'uuid2': AttributeInfo(name='sex')},
                                predicates={})
    assert await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testGePredicate(prover1, verifier, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 18)})
    assert await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testGePredicateForEqual(prover1, verifier, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 28)})
    assert await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testGePredicateNegative(prover1, verifier, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid': PredicateGE('age', 29)})
    with pytest.raises(ValueError):
        await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testMultipleGePredicate(prover1, verifier, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('height', 170)})
    assert await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testMultipleGePredicateNegative(prover1, verifier, claimsProver1Gvt):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('height', 180)})
    with pytest.raises(ValueError):
        await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testNonceShouldBeSame(prover1, verifier, claimsProver1Gvt, nonce,
                                genNonce):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={'attr_uuid': AttributeInfo(name='name')})

    proof = await prover1.presentProof(proofRequest)

    proofRequest = ProofRequest("proof1", "1.0", genNonce,
                                verifiableAttributes=proofRequest.verifiableAttributes,
                                predicates=proofRequest.predicates)
    assert not await verifier.verify(proofRequest, proof)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testUParamShouldBeSame(prover1, verifier, issuerGvt, schemaGvtId,
                                 attrsProver1Gvt, keysGvt,
                                 issueAccumulatorGvt):
    claimsReq = await prover1.createClaimRequest(schemaGvtId)

    claimsReq = claimsReq._replace(U=claimsReq.U ** 2)
    claim_signature, claim_attributes = await issuerGvt.issueClaim(schemaGvtId, claimsReq)
    await prover1.processClaim(schemaGvtId, claim_attributes, claim_signature)

    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid': AttributeInfo(name='name')},
                                predicates={})
    assert not await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.asyncio
async def testUrParamShouldBeSame(prover1, issuerGvt, schemaGvtId,
                                  attrsProver1Gvt, keysGvt,
                                  issueAccumulatorGvt):
    claimsReq = await prover1.createClaimRequest(schemaGvtId)

    claimsReq = claimsReq._replace(Ur=claimsReq.Ur ** 2)
    claim_signature, claim_attributes = await issuerGvt.issueClaim(schemaGvtId, claimsReq)

    with pytest.raises(ValueError):
        await prover1.processClaim(schemaGvtId, claim_attributes, claim_signature)



import pytest


def run():
    runs = 40
    fails = 0
    for x in range(runs):
        itr = x + 1
        print('running {}'.format(itr))
        # tst = 'test_greater_eq_predicate.py::testPredicateGreaterEqMultiIssuers'
        # result = pytest.main('-x --tb=long -n 7 {}'.format(tst))
        result = pytest.main('-x')
        failed = bool(result)
        fails += int(failed)
    print("{} runs, {} failures".format(runs, fails))


if __name__ == '__main__':
    run()



import pytest

from anoncreds.protocol.primary.primary_proof_builder import fourSquares


def testQuadEquationLagranges():
    delta = 85
    u = fourSquares(delta)
    print("u1: {0} u2: {1} u3: {2} u4: {3}".format(u['0'], u['1'], u['2'],
                                                   u['3']))
    assert (u['0'] ** 2) + (u['1'] ** 2) + (u['2'] ** 2) + (
        u['3'] ** 2) == delta


def testQuadEquationLagrangesNegativeInt():
    delta = -5
    with pytest.raises(ValueError):
        fourSquares(delta)



import os
from collections import namedtuple
from typing import Sequence, Set, TypeVar

from anoncreds.protocol.utils import toDictWithStrValues, \
    fromDictWithStrValues, encodeAttr, crypto_int_to_str, to_crypto_int, isCryptoInteger, \
    intToArrayBytes, bytesToInt
from config.config import cmod
from typing import NamedTuple
import uuid


class AttribType:
    def __init__(self, name: str, encode: bool):
        self.name = name
        self.encode = encode

    def __eq__(self, y):
        return self.__dict__ == y.__dict__

    def __lt__(self, other):
        return self.name < other.name

    def __repr__(self):
        return str(self.__dict__)


class AttribDef:
    def __init__(self, name, attrTypes):
        if isinstance(name, str):
            self.names = [name]
            self.attrTypes = [attrTypes]
        else:
            self.names = name
            self.attrTypes = attrTypes

    @property
    def name(self):
        return ', '.join(self.names)

    def __getattr__(self, item):
        for attr_types in self.attrTypes:
            for at in attr_types:
                if item == at.name:
                    return at
            raise AttributeError

    def __add__(self, other):
        return AttribDef(self.names + other.names,
                         self.attrTypes + other.attrTypes)

    def attribs(self, **vals):
        for k in vals:
            # Check that keys provided in vals match the attibute names
            # in the schema definition
            assert k in self.attribNames()
        return Attribs(self, **vals)

    def attribNames(self):
        return [at.name
                for attr_types in self.attrTypes
                for at in attr_types]

    def __eq__(self, y):
        return sorted(self.names) == sorted(y.names) \
            and sorted(self.attrTypes) == sorted(y.attrTypes)

    def __repr__(self):
        return str(self.__dict__)


class Attribs:
    def __init__(self, credType: AttribDef = None, **vals):
        self.credType = credType if credType else AttribDef([], [])
        self._vals = vals

    def encoded(self):
        """
        This function will encode all the attributes to 256 bit integers

        :return:
        """

        encoded = {}
        for i in range(len(self.credType.names)):
            self.credType.names[i]
            attr_types = self.credType.attrTypes[i]

            for at in attr_types:
                attrName = at.name
                if attrName in self._vals:
                    if at.encode:
                        encoded[attrName] = encodeAttr(self._vals[attrName])
                    else:
                        encoded[attrName] = self._vals[at.name]
        return encoded

    def __add__(self, other):
        vals = self._vals.copy()
        vals.update(other._vals)
        return Attribs(self.credType + other.credType, **vals)

    def __iter__(self):
        return self._vals.__iter__()

    def __getitem__(self, key):
        return self._vals[key]

    def keys(self):
        return self._vals.keys()

    def values(self):
        return self._vals.values()

    def items(self):
        return self._vals.items()

    def __repr__(self):
        return str(self.__dict__)

    def __eq__(self, y):
        return self.credType == y.credType \
            and self._vals == y._vals


PublicParams = namedtuple('PublicParams', 'Gamma, rho, g, h')

T = TypeVar('T')
VType = Set[int]
TimestampType = int


class Tails:

    def __init__(self):
        self.g = {}
        self.gprime = {}

    def addValue(self, index, gVal, gprimeVal):
        self.g[index] = gVal
        self.gprime[index] = gprimeVal


class NamedTupleStrSerializer:
    def toStrDict(self):
        return toDictWithStrValues(self._asdict())

    @classmethod
    def fromStrDict(cls, d):
        d = fromDictWithStrValues(d)
        return cls(**d)


class StrSerializer:
    def toStrDict(self):
        return toDictWithStrValues(self.__dict__)

    @classmethod
    def fromStrDict(cls, d):
        d = fromDictWithStrValues(d)
        return cls(**d)


class SchemaKey(
        namedtuple('SchemaKey', 'name, version, issuerId'),
        NamedTupleStrSerializer):
    def __new__(cls, name=None, version=None, issuerId=None):
        return super(SchemaKey, cls).__new__(cls, name, version, issuerId)

    def __hash__(self):
        keys = (self.name, self.version, self.issuerId)
        return hash(keys)

    def __str__(self):
        rtn = list()
        rtn.append('Schema Key')
        rtn.append("    Name: {}".format(str(self.name)))
        rtn.append("    Version: {}".format(str(self.version)))
        rtn.append("    IssuerId: {}".format(str(self.issuerId)))

        return os.linesep.join(rtn)


class ID(namedtuple('ID', 'schemaKey, schemaId, seqId')):
    def __new__(cls, schemaKey: SchemaKey = None, schemaId=None,
                seqId=None):
        return super(ID, cls).__new__(cls, schemaKey, schemaId, seqId)


class Schema(namedtuple('Schema',
                        'name, version, attrNames, issuerId, seqId'),
             NamedTupleStrSerializer):
    def __new__(cls, name, version, attrNames, issuerId, seqId=None):
        return super(Schema, cls).__new__(cls,
                                          name,
                                          version,
                                          attrNames,
                                          issuerId,
                                          seqId)

    def getKey(self):
        return SchemaKey(self.name, self.version, self.issuerId)


class PublicKey(namedtuple('PublicKey', 'N, Rms, Rctxt, R, S, Z, seqId'),
                NamedTupleStrSerializer):
    def __new__(cls, N, Rms, Rctxt, R, S, Z, seqId=None):
        return super(PublicKey, cls).__new__(cls, N, Rms, Rctxt, R, S, Z, seqId)

    def __eq__(self, other):
        return self.N == other.N and self.Rms == other.Rms \
            and self.Rctxt == other.Rctxt and self.S == other.S \
            and self.Z == other.Z and self.seqId == other.seqId \
            and dict(self.R) == dict(other.R)

    def to_str_dict(self):
        public_key = {
            'n': str(crypto_int_to_str(self.N)),
            's': str(crypto_int_to_str(self.S)),
            'rms': str(crypto_int_to_str(self.Rms)),
            'rctxt': str(crypto_int_to_str(self.Rctxt)),
            'z': str(crypto_int_to_str(self.Z)),
            'r': {k: str(crypto_int_to_str(v)) for k, v in self.R.items()}
        }

        return public_key

    @classmethod
    def from_str_dict(cls, data):
        N = to_crypto_int(data['n'])
        Rms = to_crypto_int(data['rms'], data['n'])
        Rctxt = to_crypto_int(data['rctxt'], data['n'])
        S = to_crypto_int(data['s'], data['n'])
        Z = to_crypto_int(data['z'], data['n'])
        R = {k: to_crypto_int(v, data['n']) for k, v in data['r'].items()}

        return cls(N, Rms, Rctxt, R, S, Z)


class SecretKey(namedtuple('SecretKey', 'pPrime, qPrime'),
                NamedTupleStrSerializer):
    pass


class RevocationPublicKey(namedtuple('RevocationPublicKey',
                                     'qr, g, gprime, h, h0, h1, h2, htilde, hhat, u, pk, y, seqId'),
                          NamedTupleStrSerializer):
    def __new__(cls, qr, g, gprime, h, h0, h1, h2, htilde, hhat, u, pk, y, seqId=None):
        return super(RevocationPublicKey, cls).__new__(cls, qr, g, gprime, h, h0, h1,
                                                       h2, htilde, hhat, u, pk, y,
                                                       seqId)


class RevocationSecretKey(namedtuple('RevocationSecretKey', 'x, sk'),
                          NamedTupleStrSerializer):
    pass


class AccumulatorPublicKey(namedtuple('AccumulatorPublicKey', 'z, seqId'),
                           NamedTupleStrSerializer):
    def __new__(cls, z, seqId=None):
        return super(AccumulatorPublicKey, cls).__new__(cls, z, seqId)


class AccumulatorSecretKey(
        namedtuple('AccumulatorSecretKey', 'gamma'), NamedTupleStrSerializer):
    pass


class Predicate(namedtuple('Predicate', 'attrName, value, type, schema_seq_no, issuer_did'),
                NamedTupleStrSerializer):
    def __new__(cls, attrName, value, type, schema_seq_no=None, issuer_did=None):
        return super(Predicate, cls).__new__(cls, attrName, value, type, schema_seq_no, issuer_did)

    def __key(self):
        return self.attrName, self.value, self.type

    def __eq__(self, y):
        return self.__key() == y.__key()

    def __hash__(self):
        return hash(self.__key())

    def to_str_dict(self):
        return {
            'attr_name': self.attrName,
            'value': self.value,
            'p_type': self.type,
            'schema_seq_no': self.schema_seq_no,
            'issuer_did': self.issuer_did
        }

    @classmethod
    def from_str_dict(cls, d):
        attrName = d['attr_name']
        value = d['value']
        type = d['p_type']
        schema_seq_no = int(d['schema_seq_no']) if (
            ('schema_seq_no' in d) and d['schema_seq_no']) else None
        issuer_did = int(d['issuer_did']) if (
            ('issuer_did' in d) and d['issuer_did']) else None
        return PredicateGE(attrName=attrName, value=value, type=type,
                           schema_seq_no=schema_seq_no, issuer_did=issuer_did)


# TODO: now we consdider only  >= predicate. Support other types of predicates
class PredicateGE(Predicate):
    def __new__(cls, attrName, value, type='GE', schema_seq_no=None, issuer_did=None):
        return super(PredicateGE, cls).__new__(cls, attrName, value, type, schema_seq_no, issuer_did)


class Accumulator:
    def __init__(self, iA, acc, V: VType, L):
        self.iA = iA
        self.acc = acc
        self.V = V
        self.L = L
        self.currentI = 1

    def isFull(self):
        return self.currentI > self.L

    def __eq__(self, other):
        return self.iA == other.iA and self.acc == other.acc \
            and self.V == other.V and self.L == other.L \
            and self.currentI == other.currentI


ClaimInitDataType = namedtuple('ClaimInitDataType', 'U, vPrime')


class ClaimRequest(namedtuple('ClaimRequest', 'userId, U, Ur'),
                   NamedTupleStrSerializer):
    def __new__(cls, userId, U, Ur=None):
        return super(ClaimRequest, cls).__new__(cls, userId, U, Ur)

    def to_str_dict(self):
        return {
            'prover_did': str(self.userId),
            'u': str(crypto_int_to_str(self.U)),
            'ur': self.Ur
        }

    @classmethod
    def from_str_dict(cls, data, n):
        u = to_crypto_int(data['u'], str(n))

        return cls(userId=data['prover_did'], U=u, Ur=data['ur'])


# Accumulator = namedtuple('Accumulator', ['iA', 'acc', 'V', 'L'])

class PrimaryClaim(
        namedtuple('PrimaryClaim', 'm2, A, e, v'),
        NamedTupleStrSerializer):

    def to_str_dict(self):
        return {
            'm2': str(crypto_int_to_str(self.m2)),
            'a': str(crypto_int_to_str(self.A)),
            'e': str(self.e),
            'v': str(self.v)
        }

    @classmethod
    def from_str_dict(cls, data, n):
        m2 = to_crypto_int(data['m2'])
        a = to_crypto_int(data['a'], str(n))
        e = int(data['e'])
        v = int(data['v'])

        return cls(m2=m2, A=a, e=e, v=v)


class Witness(namedtuple('Witness', 'sigmai, ui, gi, omega, V'),
              NamedTupleStrSerializer):
    pass


class NonRevocationClaim(
        namedtuple('NonRevocationClaim', 'iA, sigma, c, v, witness,i, m2'),
        NamedTupleStrSerializer):
    @classmethod
    def fromStrDict(cls, d):
        d = fromDictWithStrValues(d)
        witness = Witness(**d['witness'])
        result = cls(**d)
        return result._replace(witness=witness)

    def to_str_dict(self):
        return {
        }


class Claims(namedtuple('Claims', 'primaryClaim, nonRevocClaim'),
             NamedTupleStrSerializer):
    def __new__(cls, primaryClaim, nonRevocClaim=None):
        return super(Claims, cls).__new__(cls, primaryClaim, nonRevocClaim)

    @classmethod
    def fromStrDict(cls, d):
        primary = PrimaryClaim.fromStrDict(d['primaryClaim'])
        nonRevoc = None
        if 'nonRevocClaim' in d:
            nonRevoc = NonRevocationClaim.fromStrDict(d['nonRevocClaim'])
        return Claims(primaryClaim=primary, nonRevocClaim=nonRevoc)

    def to_str_dict(self):
        return {
            'primary_claim': self.primaryClaim.to_str_dict(),
            'non_revocation_claim': self.nonRevocClaim.to_str_dict() if self.nonRevocClaim else None
        }

    @classmethod
    def from_str_dict(cls, data, n):
        primary = PrimaryClaim.from_str_dict(data['primary_claim'], n)
        nonRevoc = None
        if 'non_revocation_claim' in data and data['non_revocation_claim']:
            nonRevoc = NonRevocationClaim.fromStrDict(
                data['non_revocation_claim'])

        return cls(primaryClaim=primary, nonRevocClaim=nonRevoc)


class ClaimsPair(dict):
    def __str__(self):
        rtn = list()
        rtn.append('Claims')

        for schema_key, claim_attrs in self.items():
            rtn.append('')
            rtn.append(schema_key.name)
            rtn.append(str(schema_key))
            rtn.append('Attributes:')
            for attr_name, attr_raw_enc in claim_attrs.items():
                rtn.append('    {}: {}'.format(str(attr_name),
                                               str(attr_raw_enc)))

        return os.linesep.join(rtn)


class AttributeInfo(
        namedtuple('AttributeInfo', 'name, schema_seq_no, issuer_did'),
        NamedTupleStrSerializer):
    def __new__(cls, name=None, schema_seq_no=None, issuer_did=None):
        return super(AttributeInfo, cls).__new__(cls, name, schema_seq_no, issuer_did)

    def to_str_dict(self):
        return {
            'name': self.name,
            'schema_seq_no': self.schema_seq_no,
            'issuer_did': self.issuer_did
        }

    @classmethod
    def from_str_dict(cls, d):
        schema_seq_no = int(d['schema_seq_no']) if d['schema_seq_no'] else None
        issuer_did = int(d['issuer_did']) if (
            ('issuer_did' in d) and d['issuer_did']) else None
        name = d['name']
        return AttributeInfo(name, schema_seq_no, issuer_did)


class ProofClaims(
        namedtuple('ProofClaims', 'claims, revealedAttrs, predicates')):
    def __new__(cls, claims=None, revealedAttrs=None, predicates=None):
        return super(ProofClaims, cls).__new__(cls, claims, revealedAttrs or [],
                                               predicates or [])


class NonRevocProofXList(
        namedtuple('NonRevocProofXList',
                   'rho, r, rPrime, rPrimePrime, rPrimePrimePrime, o, oPrime, m, mPrime, t, tPrime, m2, s, c'),
        NamedTupleStrSerializer):
    def __new__(cls, rho=None, r=None, rPrime=None, rPrimePrime=None,
                rPrimePrimePrime=None, o=None, oPrime=None,
                m=None, mPrime=None, t=None, tPrime=None, m2=None, s=None,
                c=None, group=None):
        return super(NonRevocProofXList, cls).__new__(cls,
                                                      rho=cls._setValue(rho,
                                                                        group),
                                                      r=cls._setValue(
                                                          r, group),
                                                      rPrime=cls._setValue(
                                                          rPrime, group),
                                                      rPrimePrime=cls._setValue(
                                                          rPrimePrime, group),
                                                      rPrimePrimePrime=cls._setValue(
                                                          rPrimePrimePrime,
                                                          group),
                                                      o=cls._setValue(
                                                          o, group),
                                                      oPrime=cls._setValue(
                                                          oPrime, group),
                                                      m=cls._setValue(
                                                          m, group),
                                                      mPrime=cls._setValue(
                                                          mPrime, group),
                                                      t=cls._setValue(
                                                          t, group),
                                                      tPrime=cls._setValue(
                                                          tPrime, group),
                                                      m2=cls._setValue(m2,
                                                                       group),
                                                      s=cls._setValue(
                                                          s, group),
                                                      c=cls._setValue(c, group))

    @staticmethod
    def _setValue(v=None, group=None):
        return v if v else group.random(cmod.ZR) if group else None

    def asList(self):
        return [self.rho, self.o, self.c, self.oPrime, self.m, self.mPrime,
                self.t, self.tPrime,
                self.m2, self.s, self.r, self.rPrime, self.rPrimePrime,
                self.rPrimePrimePrime]

    @staticmethod
    def fromList(values: Sequence):
        rho, o, c, oPrime, m, mPrime, t, tPrime, m2, s, r, rPrime, rPrimePrime, rPrimePrimePrime = tuple(
            values)
        return NonRevocProofXList(rho=rho, o=o, c=c, oPrime=oPrime, m=m,
                                  mPrime=mPrime, t=t, tPrime=tPrime,
                                  m2=m2, s=s, r=r, rPrime=rPrime,
                                  rPrimePrime=rPrimePrime,
                                  rPrimePrimePrime=rPrimePrimePrime)


class NonRevocProofCList(
        namedtuple('NonRevocProofCList', 'E, D, A, G, W, S, U'),
        NamedTupleStrSerializer):
    def asList(self):
        return [self.E, self.D, self.A, self.G, self.W, self.S, self.U]


class NonRevocProofTauList(
        namedtuple('NonRevocProofTauList', 'T1, T2, T3, T4, T5, T6, T7, T8'),
        NamedTupleStrSerializer):
    def asList(self):
        return [self.T1, self.T2, self.T3, self.T4, self.T5, self.T6, self.T7,
                self.T8]


class NonRevocInitProof(namedtuple('NonRevocInitProof',
                                   'CList, TauList, CListParams, TauListParams'),
                        NamedTupleStrSerializer):
    def asCList(self):
        return self.CList.asList()

    def asTauList(self):
        return self.TauList.asList()


class PrimaryEqualInitProof(namedtuple('PrimaryEqualInitProof',
                                       'c1, Aprime, T, eTilde, ePrime, vTilde, vPrime, \
                                       mTilde, m1Tilde, m2Tilde, unrevealedAttrs, revealedAttrs'),
                            NamedTupleStrSerializer):
    def asCList(self):
        return [self.Aprime]

    def asTauList(self):
        return [self.T]


class PrimaryPrecicateGEInitProof(
        namedtuple('PrimaryPrecicateGEInitProof',
                   'CList, TauList, u, uTilde, r, rTilde, alphaTilde, predicate, T'),
        NamedTupleStrSerializer):
    def asCList(self):
        return self.CList

    def asTauList(self):
        return self.TauList


class PrimaryInitProof(namedtuple('PrimaryInitProof', 'eqProof, geProofs'),
                       NamedTupleStrSerializer):
    def asCList(self):
        CList = self.eqProof.asCList()
        for geProof in self.geProofs:
            CList += geProof.asCList()
        return CList

    def asTauList(self):
        TauList = self.eqProof.asTauList()
        for geProof in self.geProofs:
            TauList += geProof.asTauList()
        return TauList


class InitProof(namedtuple('InitProof', 'nonRevocInitProof, primaryInitProof'),
                NamedTupleStrSerializer):
    def __new__(cls, nonRevocInitProof: NonRevocInitProof = None,
                primaryInitProof: PrimaryInitProof = None):
        return super(InitProof, cls).__new__(cls, nonRevocInitProof,
                                             primaryInitProof)


class PrimaryEqualProof(namedtuple('PrimaryEqualProof',
                                   'e, v, m, m1, m2, Aprime, revealedAttrs'),
                        NamedTupleStrSerializer):

    def to_str_dict(self):
        return {
            'e': str(crypto_int_to_str(self.e)),
            'v': str(crypto_int_to_str(self.v)),
            'm1': str(crypto_int_to_str(self.m1)),
            'm2': str(crypto_int_to_str(self.m2)),
            'm': {k: str(crypto_int_to_str(v)) for k, v in self.m.items()},
            'revealed_attrs': {k: str(v) for k, v in self.revealedAttrs.items()},
            'a_prime': str(crypto_int_to_str(self.Aprime))
        }

    @classmethod
    def from_str_dict(cls, d, n):
        e = to_crypto_int(d['e'])
        v = to_crypto_int(d['v'])
        m1 = to_crypto_int(d['m1'])
        m2 = to_crypto_int(d['m2'])
        Aprime = to_crypto_int(d['a_prime'], str(n))
        revealedAttrs = {k: to_crypto_int(v)
                         for k, v in d['revealed_attrs'].items()}
        m = {k: to_crypto_int(v) for k, v in d['m'].items()}

        return PrimaryEqualProof(e=e, v=v, m1=m1, m2=m2, m=m, Aprime=Aprime, revealedAttrs=revealedAttrs)


class PrimaryPredicateGEProof(
        namedtuple('PrimaryPredicateGEProof', 'u, r, alpha, mj, T, predicate'),
        NamedTupleStrSerializer):
    @classmethod
    def fromStrDict(cls, d):
        d = fromDictWithStrValues(d)
        predicate = PredicateGE(**d['predicate'])
        result = cls(**d)
        return result._replace(predicate=predicate)

    def to_str_dict(self):
        return {
            'alpha': str(crypto_int_to_str(self.alpha)),
            'mj': str(crypto_int_to_str(self.mj)),
            'u': {k: str(crypto_int_to_str(v)) for k, v in self.u.items()},
            'r': {k: str(crypto_int_to_str(v)) for k, v in self.r.items()},
            't': {k: str(crypto_int_to_str(v)) for k, v in self.T.items()},
            'predicate': self.predicate.to_str_dict()
        }

    @classmethod
    def from_str_dict(cls, d, n):
        alpha = to_crypto_int(d['alpha'])
        mj = to_crypto_int(d['mj'])
        u = {k: to_crypto_int(v) for k, v in d['u'].items()}
        r = {k: to_crypto_int(v) for k, v in d['r'].items()}
        T = {k: to_crypto_int(v, str(n)) for k, v in d['t'].items()}
        predicate = PredicateGE.from_str_dict(d['predicate'])

        return PrimaryPredicateGEProof(alpha=alpha, mj=mj, u=u, r=r, T=T, predicate=predicate)


class NonRevocProof(namedtuple('NonRevocProof', 'XList CProof'),
                    NamedTupleStrSerializer):
    @classmethod
    def fromStrDict(cls, d):
        XList = NonRevocProofXList.fromStrDict(d['XList'])
        CProof = NonRevocProofCList.fromStrDict(d['CProof'])
        return NonRevocProof(XList=XList, CProof=CProof)


class PrimaryProof(namedtuple('PrimaryProof', 'eqProof, geProofs'),
                   NamedTupleStrSerializer):
    def __new__(cls, eqProof: PrimaryEqualProof,
                geProofs: Sequence[PrimaryPredicateGEProof]):
        return super(PrimaryProof, cls).__new__(cls, eqProof, geProofs)

    @classmethod
    def fromStrDict(cls, d):
        eqProof = PrimaryEqualProof.fromStrDict(d['eqProof'])
        geProofs = [PrimaryPredicateGEProof.fromStrDict(
            v) for v in d['geProofs']]
        return PrimaryProof(eqProof=eqProof, geProofs=geProofs)

    def to_str_dict(self):
        return {
            'eq_proof': self.eqProof.to_str_dict(),
            'ge_proofs': [p.to_str_dict() for p in self.geProofs]
        }

    @classmethod
    def from_str_dict(cls, d, n):
        eqProof = PrimaryEqualProof.from_str_dict(d['eq_proof'], n)
        geProofs = [PrimaryPredicateGEProof.from_str_dict(
            p, n) for p in d['ge_proofs']]

        return PrimaryProof(eqProof=eqProof, geProofs=geProofs)


class Proof(namedtuple('Proof', 'primaryProof, nonRevocProof'),
            NamedTupleStrSerializer):
    def __new__(cls, primaryProof: PrimaryProof,
                nonRevocProof: NonRevocProof = None):
        return super(Proof, cls).__new__(cls, primaryProof, nonRevocProof)

    @classmethod
    def fromStrDict(cls, d):
        primaryProof = PrimaryProof.fromStrDict(d['primaryProof'])
        nonRevocProof = None
        if 'nonRevocProof' in d:
            nonRevocProof = NonRevocProof.fromStrDict(d['nonRevocProof'])
        return Proof(primaryProof=primaryProof, nonRevocProof=nonRevocProof)

    def to_str_dict(self):
        return {
            'primary_proof': self.primaryProof.to_str_dict()
        }

    @classmethod
    def from_str_dict(cls, d, n):
        primaryProof = PrimaryProof.from_str_dict(d['primary_proof'], n)

        return Proof(primaryProof=primaryProof)


class ProofInfo(namedtuple('ProofInfo', 'proof, schema_seq_no, issuer_did'),
                NamedTupleStrSerializer):
    @classmethod
    def fromStrDict(cls, d):
        d = fromDictWithStrValues(d)
        proof = Proof.fromStrDict(d['proof'])
        result = cls(**d)
        return result._replace(proof=proof)

    def to_str_dict(self):
        return {
            'proof': self.proof.to_str_dict(),
            'schema_seq_no': self.schema_seq_no,
            'issuer_did': self.issuer_did
        }

    @classmethod
    def from_str_dict(cls, d, n):
        proof = Proof.from_str_dict(d['proof'], n)
        schema_seq_no = d['schema_seq_no']
        issuer_did = d['issuer_did']

        return ProofInfo(proof=proof, schema_seq_no=schema_seq_no, issuer_did=issuer_did)


class FullProof(namedtuple('FullProof', 'proofs, aggregatedProof, requestedProof'),
                NamedTupleStrSerializer):
    def getCredDefs(self):
        return self.proofs.keys()

    @classmethod
    def fromStrDict(cls, d):
        d = fromDictWithStrValues(d)
        aggregatedProof = AggregatedProof.fromStrDict(d['aggregatedProof'])
        requestedProof = RequestedProof.fromStrDict(d['requestedProof'])
        proofs = {k: ProofInfo.fromStrDict(v) for k, v in d['proofs'].items()}

        return FullProof(aggregatedProof=aggregatedProof, proofs=proofs, requestedProof=requestedProof)

    def to_str_dict(self):
        return {
            'aggregated_proof': self.aggregatedProof.to_str_dict(),
            'proofs': {k: v.to_str_dict() for k, v in self.proofs.items()},
            'requested_proof': self.requestedProof.to_str_dict()
        }

    @classmethod
    def from_str_dict(cls, d, n):
        aggregatedProof = AggregatedProof.from_str_dict(d['aggregated_proof'])
        requestedProof = RequestedProof.from_str_dict(d['requested_proof'])
        proofs = {item[0]: ProofInfo.from_str_dict(
            item[1], n[i]) for i, item in enumerate(d['proofs'].items())}

        return FullProof(aggregatedProof=aggregatedProof, requestedProof=requestedProof, proofs=proofs)


class AggregatedProof(namedtuple('AggregatedProof', 'cHash, CList'),
                      NamedTupleStrSerializer):
    def to_str_dict(self):
        return {
            'c_hash': str(self.cHash),
            'c_list': [intToArrayBytes(v) for v in self.CList if isCryptoInteger(v)]
        }

    @classmethod
    def from_str_dict(cls, d):
        cHash = int(d['c_hash'])
        CList = [bytesToInt(v) for v in d['c_list']]
        return AggregatedProof(cHash=cHash, CList=CList)


class RequestedProof(namedtuple('RequestedProof', 'revealed_attrs, unrevealed_attrs, self_attested_attrs, predicates'),
                     NamedTupleStrSerializer):
    def __new__(cls, revealed_attrs=None, unrevealed_attrs=None, self_attested_attrs=None, predicates=None):
        return super(RequestedProof, cls).__new__(cls, revealed_attrs or {}, unrevealed_attrs or {},
                                                  self_attested_attrs or {}, predicates or {})

    @classmethod
    def fromStrDict(cls, d):
        revealed_attrs = {k: [v[0], v[1], v[2]]
                          for k, v in d['revealed_attrs'].items()}
        predicates = {k: v for k, v in d['predicates'].items()}
        return RequestedProof(revealed_attrs=revealed_attrs, predicates=predicates)

    def to_str_dict(self):
        return {
            'revealed_attrs': self.revealed_attrs,
            'unrevealed_attrs': self.unrevealed_attrs,
            'self_attested_attrs': self.self_attested_attrs,
            'predicates': self.predicates
        }

    @classmethod
    def from_str_dict(cls, d):
        revealed_attrs = d['revealed_attrs']
        unrevealed_attrs = d['unrevealed_attrs']
        self_attested_attrs = d['self_attested_attrs']
        predicates = d['predicates']
        return RequestedProof(revealed_attrs=revealed_attrs, unrevealed_attrs=unrevealed_attrs,
                              self_attested_attrs=self_attested_attrs, predicates=predicates)


class ClaimAttributeValues(namedtuple('ClaimAttributeValues', 'raw, encoded'),
                           NamedTupleStrSerializer):
    def __new__(cls, raw=None, encoded=None):
        return super(ClaimAttributeValues, cls).__new__(cls, raw, encoded)

    def __str__(self):
        return self.raw

    def to_str_dict(self):
        return [str(self.raw), str(self.encoded)]

    @classmethod
    def from_str_dict(cls, d):
        raw = d[0]
        encoded = int(to_crypto_int(d[1]))
        return ClaimAttributeValues(raw=raw, encoded=encoded)


AvailableClaim = NamedTuple("AvailableClaim", [("name", str),
                                               ("version", str),
                                               ("origin", str)])


class ProofRequest:
    def __init__(self, name, version, nonce, attributes={}, verifiableAttributes={}, predicates={}):
        self.name = name
        self.version = version
        self.nonce = nonce
        self.attributes = attributes
        self.verifiableAttributes = \
            {str(uuid.uuid4()): AttributeInfo(name=a) for a in verifiableAttributes} if \
            isinstance(verifiableAttributes, list) else verifiableAttributes
        self.predicates = {str(uuid.uuid4()): PredicateGE(attrName=p['attrName'], value=p['value']) for p in
                           predicates} if isinstance(predicates, list) else predicates
        self.fulfilledByClaims = []
        self.selfAttestedAttrs = {}
        self.ts = None
        self.seqNo = None
        # TODO _F_ need to add support for predicates on unrevealed attibutes

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    @property
    def toDict(self):
        return {
            "name": self.name,
            "version": self.version,
            "nonce": self.nonce,
            "attributes": self.attributes,
            "verifiableAttributes": self.verifiableAttributes
        }

    def to_str_dict(self):
        return {
            "name": self.name,
            "version": self.version,
            "nonce": str(self.nonce),
            "requested_attrs": {k: v.to_str_dict() for k, v in self.verifiableAttributes.items()},
            "requested_predicates": {k: v.to_str_dict() for k, v in self.predicates.items()}
        }

    @staticmethod
    def from_str_dict(d):
        return ProofRequest(name=d['name'],
                            version=d['version'],
                            nonce=int(d['nonce']),
                            attributes=d['attributes'] if 'attributes' in d else {
        },
            verifiableAttributes={k: AttributeInfo.from_str_dict(v) for k, v in
                                  d['requested_attrs'].items()},
            predicates={k: PredicateGE.from_str_dict(v) for k, v in d['requested_predicates'].items()})

    @property
    def attributeValues(self):
        return \
            'Attributes:' + '\n    ' + \
            format("\n    ".join(
                ['{}: {}'.format(k, v)
                 for k, v in self.attributes.items()])) + '\n'

    @property
    def verifiableClaimAttributeValues(self):
        return \
            'Verifiable Attributes:' + '\n    ' + \
            format("\n    ".join(
                ['{}'.format(v.name)
                 for k, v in self.verifiableAttributes.items()])) + '\n'

    @property
    def predicateValues(self):
        return \
            'Predicates:' + '\n    ' + \
            format("\n    ".join(
                ['{}'.format(v.attrName)
                 for k, v in self.predicates.items()])) + '\n'

    @property
    def fixedInfo(self):
        return 'Status: Requested' + '\n' + \
               'Name: ' + self.name + '\n' + \
               'Version: ' + self.version + '\n'

    def __str__(self):
        return 'Proof Request\n' + \
               self.fixedInfo + \
               self.attributeValues + \
               self.verifiableClaimAttributeValues



import logging
import string
import time
from collections import OrderedDict
from enum import Enum
from hashlib import sha256
from math import sqrt, floor
from random import randint, sample
from sys import byteorder
from typing import Dict, List, Set

import base58

from anoncreds.protocol.globals import KEYS, PK_R
from anoncreds.protocol.globals import LARGE_PRIME, LARGE_MASTER_SECRET, \
    LARGE_VPRIME, PAIRING_GROUP
from config.config import cmod
import sys


def encodeAttr(attrValue):
    return cmod.Conversion.bytes2integer(sha256(str(attrValue).encode()).digest())


def randomQR(n):
    return cmod.random(n) ** 2


def get_hash_as_int(*args, group: cmod.PairingGroup = None):
    """
    Enumerate over the input tuple and generate a hash using the tuple values

    :param args: sequence of either group or integer elements
    :param group: pairing group if an element is a group element
    :return:
    """

    group = group if group else cmod.PairingGroup(PAIRING_GROUP)
    h_challenge = sha256()

    serialedArgs = [group.serialize(arg) if isGroupElement(arg)
                    else cmod.Conversion.IP2OS(arg)
                    for arg in args]

    for arg in sorted(serialedArgs):
        h_challenge.update(arg)
    return bytes_to_int(h_challenge.digest())


CRYPTO_INT_PREFIX = 'CryptoInt_'
INT_PREFIX = 'Int_'
GROUP_PREFIX = 'Group_'
BYTES_PREFIX = 'Bytes_'


def serializeToStr(n):
    if isCryptoInteger(n):
        return CRYPTO_INT_PREFIX + cmod.serialize(n).decode()
    if isInteger(n):
        return INT_PREFIX + str(n)
    if isGroupElement(n):
        return GROUP_PREFIX + cmod.PairingGroup(PAIRING_GROUP).serialize(
            n).decode()
    return n


def deserializeFromStr(n: str):
    if isStr(n) and n.startswith(CRYPTO_INT_PREFIX):
        n = n[len(CRYPTO_INT_PREFIX):].encode()
        return cmod.deserialize(n)

    if isStr(n) and n.startswith(INT_PREFIX):
        n = n[len(INT_PREFIX):]
        return int(n)

    if isStr(n) and n.startswith(GROUP_PREFIX):
        n = n[len(GROUP_PREFIX):].encode()
        res = cmod.PairingGroup(PAIRING_GROUP).deserialize(n)
        # A fix for Identity element as serialized/deserialized not correctly
        if str(res) == '[0, 0]':
            return groupIdentityG1()
        return res

    return n


def isCryptoInteger(n):
    return isinstance(n, cmod.integer)


def isGroupElement(n):
    return isinstance(n, cmod.pc_element)


def isInteger(n):
    return isinstance(n, int)


def isStr(n):
    return isinstance(n, str)


def isNamedTuple(n):
    return isinstance(n, tuple)  # TODO: assume it's a named tuple


def toDictWithStrValues(d):
    if isNamedTuple(d):
        return toDictWithStrValues(d._asdict())
    if not isinstance(d, Dict):
        return serializeToStr(d)
    result = OrderedDict()
    for key, value in d.items():
        if isinstance(value, Dict):
            result[serializeToStr(key)] = toDictWithStrValues(value)
        elif isinstance(value, str):
            result[serializeToStr(key)] = serializeToStr(value)
        elif isNamedTuple(value):
            result[serializeToStr(key)] = toDictWithStrValues(value._asdict())
        elif isinstance(value, Set):
            result[serializeToStr(key)] = {toDictWithStrValues(v) for v in
                                           value}
        elif isinstance(value, List):
            result[serializeToStr(key)] = [toDictWithStrValues(v) for v in
                                           value]
        elif value:
            result[serializeToStr(key)] = serializeToStr(value)
    return result


def fromDictWithStrValues(d):
    if not isinstance(d, Dict) and not isinstance(d, tuple):
        return deserializeFromStr(d)
    result = OrderedDict()
    for key, value in d.items():
        if isinstance(value, Dict):
            result[deserializeFromStr(key)] = fromDictWithStrValues(value)
        elif isinstance(value, str):
            result[deserializeFromStr(key)] = deserializeFromStr(value)
        elif isinstance(value, Set):
            result[deserializeFromStr(key)] = {fromDictWithStrValues(v) for v in
                                               value}
        elif isinstance(value, List):
            result[deserializeFromStr(key)] = [fromDictWithStrValues(v) for v in
                                               value]
        elif value:
            result[deserializeFromStr(key)] = deserializeFromStr(value)
    return result


def bytes_to_int(bytesHash):
    return int.from_bytes(bytesHash, byteorder=byteorder)


def int_to_ZR(intHash, group):
    return group.init(cmod.ZR, intHash)


def groupIdentityG1():
    return cmod.PairingGroup(PAIRING_GROUP).init(cmod.G1, 0)


def groupIdentityG2():
    return cmod.PairingGroup(PAIRING_GROUP).init(cmod.G2, 0)


def get_values_of_dicts(*args):
    l = list()
    for d in args:
        l.extend(list(d.values()))
    return l


def get_prime_in_range(start, end):
    n = 0
    maxIter = 100000
    while n < maxIter:
        r = randint(start, end)
        if cmod.isPrime(r):
            logging.debug("Found prime in {} iterations".format(n))
            return r
        n += 1
    raise Exception("Cannot find prime in {} iterations".format(maxIter))


def splitRevealedAttrs(encodedAttrs, revealedAttrs):
    # Revealed attributes
    Ar = {}
    # Unrevealed attributes
    Aur = {}

    for k, value in encodedAttrs.items():
        if k in revealedAttrs:
            Ar[k] = value.encoded
        else:
            Aur[k] = value.encoded
    return Ar, Aur


def randomString(size: int = 20,
                 chars: str = string.ascii_letters + string.digits) -> str:
    """
    Generate a random string of the specified size.

    Ensure that the size is less than the length of chars as this function uses random.choice
    which uses random sampling without replacement.

    :param size: size of the random string to generate
    :param chars: the set of characters to use to generate the random string. Uses alphanumerics by default.
    :return: the random string generated
    """

    return ''.join(sample(chars, size))


def getUnrevealedAttrs(encodedAttrs, revealedAttrsList):
    revealedAttrs, unrevealedAttrs = splitRevealedAttrs(encodedAttrs,
                                                        revealedAttrsList)
    return unrevealedAttrs


def flattenDict(attrs):
    return {x: y for z in attrs.values()
            for x, y in z.items()}


def largestSquareLessThan(x: int):
    sqrtx = int(floor(sqrt(x)))
    return sqrtx


def fourSquares(delta: int):
    u1 = largestSquareLessThan(delta)
    u2 = largestSquareLessThan(delta - (u1 ** 2))
    u3 = largestSquareLessThan(delta - (u1 ** 2) - (u2 ** 2))
    u4 = largestSquareLessThan(delta - (u1 ** 2) - (u2 ** 2) - (u3 ** 2))
    if (u1 ** 2) + (u2 ** 2) + (u3 ** 2) + (u4 ** 2) == delta:
        return {'0': u1, '1': u2, '2': u3, '3': u4}
    else:
        raise Exception(
            "Cannot get the four squares for delta {0}".format(delta))


def strToCryptoInteger(n):
    if "mod" in n:
        a, b = n.split("mod")
        return cmod.integer(int(a.strip())) % cmod.integer(int(b.strip()))
    else:
        return cmod.integer(int(n))


def to_crypto_int(a, b=None):
    return strToCryptoInteger(a + 'mod' + b) if b else strToCryptoInteger(a)


def crypto_int_to_str(n):
    return cmod.toInt(n)


def strToInt(s):
    return bytes_to_int(sha256(s.encode()).digest())


def genPrime():
    """
    Generate 2 large primes `p_prime` and `q_prime` and use them
    to generate another 2 primes `p` and `q` of 1024 bits
    """
    prime = cmod.randomPrime(LARGE_PRIME)
    i = 0
    while not cmod.isPrime(2 * prime + 1):
        prime = cmod.randomPrime(LARGE_PRIME)
        i += 1
    return prime


def base58encode(i):
    return base58.b58encode(str(i).encode())


def base58decode(i):
    return base58.b58decode(str(i)).decode()


def base58decodedInt(i):
    try:
        return int(base58.b58decode(str(i)).decode())
    except Exception as ex:
        raise AttributeError from ex


class SerFmt(Enum):
    default = 1
    py3Int = 2
    base58 = 3


SerFuncs = {
    SerFmt.py3Int: int,
    SerFmt.default: cmod.integer,
    SerFmt.base58: base58encode,
}


def serialize(data, serFmt):
    serfunc = SerFuncs[serFmt]
    if KEYS in data:
        for k, v in data[KEYS].items():
            if isinstance(v, cmod.integer):
                # int casting works with Python 3 only.
                # for Python 2, charm's serialization api must be used.
                data[KEYS][k] = serfunc(v)
            if k == PK_R:
                data[KEYS][k] = {key: serfunc(val) for key, val in v.items()}
    return data


def generateMasterSecret():
    # Generate the master secret
    return cmod.integer(
        cmod.randomBits(LARGE_MASTER_SECRET))


def generateVPrime():
    return cmod.randomBits(LARGE_VPRIME)


def shorten(s, size=None):
    size = size or 10
    if isinstance(s, str):
        if len(s) <= size:
            return s
        else:
            head = int((size - 2) * 5 / 8)
            tail = int(size) - 2 - head
            return s[:head] + '..' + s[-tail:]
    else:  # assume it's an iterable
        return [shorten(x, size) for x in iter(s)]


def shortenMod(s, size=None):
    return ' mod '.join(shorten(str(s).split(' mod '), size))


def shortenDictVals(d, size=None):
    r = {}
    for k, v in d.items():
        if isinstance(v, dict):
            r[k] = shortenDictVals(v, size)
        else:
            r[k] = shortenMod(v, size)
    return r


def currentTimestampMillisec():
    return int(time.time() * 1000)  # millisec


def intToArrayBytes(value):
    value = int(value)
    result = []
    for i in range(0, sys.getsizeof(value)):
        b = value >> (i * 8) & 0xff
        result.append(b)

    result.reverse()

    first_non_zero = next((i for i, x in enumerate(result) if x), None)
    result = result[first_non_zero::]

    return result


def bytesToInt(bytes):
    result = 0

    for b in bytes:
        result = result * 256 + int(b)

    return result



import pytest

from anoncreds.protocol.types import ProofRequest, PredicateGE, AttributeInfo
from anoncreds.test.conftest import presentProofAndVerify


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testNoPredicates(prover1, prover2, verifier, allClaims):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={'attr_uuid1': AttributeInfo(name='name'),
                                                      'attr_uuid2': AttributeInfo(name='name')})

    assert await presentProofAndVerify(verifier, proofRequest, prover1)
    assert await presentProofAndVerify(verifier, proofRequest, prover2)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testGePredicate(prover1, prover2, verifier, allClaims):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid1': AttributeInfo(name='name')},
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('period', 3)})
    assert await presentProofAndVerify(verifier, proofRequest, prover1)
    assert await presentProofAndVerify(verifier, proofRequest, prover2)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testGePredicateNegativeForOne(prover1, prover2, verifier, allClaims):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid1': AttributeInfo(name='name')},
                                predicates={'predicate_uuid1': PredicateGE('age', 18),
                                            'predicate_uuid2': PredicateGE('period', 9)})
    assert await presentProofAndVerify(verifier, proofRequest, prover2)
    with pytest.raises(ValueError):
        await presentProofAndVerify(verifier, proofRequest, prover1)


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
@pytest.mark.asyncio
async def testGePredicateNegativeForBoth(prover1, prover2, verifier, allClaims):
    proofRequest = ProofRequest("proof1", "1.0", verifier.generateNonce(),
                                verifiableAttributes={
                                    'attr_uuid1': AttributeInfo(name='name')},
                                predicates={'predicate_uuid1': PredicateGE('age', 38),
                                            'predicate_uuid2': PredicateGE('period', 30)})
    with pytest.raises(ValueError):
        await presentProofAndVerify(verifier, proofRequest, prover1)
    with pytest.raises(ValueError):
        await presentProofAndVerify(verifier, proofRequest, prover2)



import glob
import importlib
import os
import re


def dynamicModule():
    pkg = os.path.dirname(os.path.realpath(__file__))
    clib_pattern = os.path.join(pkg, '**', 'config-crypto-*.py')
    clibs = [fn for fn in glob.glob(clib_pattern, recursive=True)]

    if len(clibs) == 0:
        raise RuntimeError('A crypto-lib must exist for anoncreds to run.')
    elif len(clibs) > 1:
        print('Found more than one crypto library, picking the first one...')
    print('Loading module {}'.format(clibs[0]))

    m = re.match('.*(config-crypto-.*)\.py', clibs[0]).groups()[0]

    mod = importlib.import_module('config.' + m)
    print('Module loaded.')
    return mod


cmod = dynamicModule()



from anoncreds.protocol.globals import PAIRING_GROUP
from anoncreds.protocol.revocation.accumulators.non_revocation_common import \
    createTauListValues, \
    createTauListExpectedValues
from anoncreds.protocol.types import NonRevocationClaim, NonRevocInitProof, \
    NonRevocProofXList, NonRevocProofCList, NonRevocProof, \
    ID, ClaimInitDataType
from anoncreds.protocol.utils import int_to_ZR
from anoncreds.protocol.wallet.prover_wallet import ProverWallet
from config.config import cmod


class NonRevocationClaimInitializer:
    def __init__(self, wallet: ProverWallet):
        self._wallet = wallet

    async def genClaimInitData(self, schemaId: ID) -> ClaimInitDataType:
        group = cmod.PairingGroup(
            PAIRING_GROUP)  # super singular curve, 1024 bits
        pkR = await self._wallet.getPublicKeyRevocation(schemaId)

        vrPrime = group.random(cmod.ZR)
        Ur = (pkR.h2 ** vrPrime)

        return ClaimInitDataType(U=Ur, vPrime=vrPrime)

    async def initNonRevocationClaim(self, schemaId: ID,
                                     claim: NonRevocationClaim):
        vrPrime = (
            await self._wallet.getNonRevocClaimInitData(schemaId)).vPrime
        newV = claim.v + vrPrime
        claim = claim._replace(v=newV)
        await self._testWitnessCredential(schemaId, claim)
        return claim

    async def _testWitnessCredential(self, schemaid: ID,
                                     claim: NonRevocationClaim):
        pkR = await self._wallet.getPublicKeyRevocation(schemaid)
        acc = await self._wallet.getAccumulator(schemaid)
        accPk = await self._wallet.getPublicKeyAccumulator(schemaid)
        m2 = int(await self._wallet.getContextAttr(schemaid))

        zCalc = cmod.pair(claim.witness.gi, acc.acc) / cmod.pair(pkR.g,
                                                                 claim.witness.omega)
        if zCalc != accPk.z:
            raise ValueError("issuer is sending incorrect data")

        pairGGCalc = cmod.pair(pkR.pk * claim.witness.gi, claim.witness.sigmai)
        pairGG = cmod.pair(pkR.g, pkR.gprime)
        if pairGGCalc != pairGG:
            raise ValueError("issuer is sending incorrect data")

        pairH1 = cmod.pair(claim.sigma, pkR.y * (pkR.hhat ** claim.c))
        pairH2 = cmod.pair(
            pkR.h0 * (pkR.h1 ** m2) * (pkR.h2 ** claim.v) * claim.witness.gi, pkR.hhat)
        if pairH1 != pairH2:
            raise ValueError("issuer is sending incorrect data")

        return True


class NonRevocationProofBuilder:
    def __init__(self, wallet: ProverWallet):
        self._wallet = wallet

    async def updateNonRevocationClaim(self, schemaId,
                                       c2: NonRevocationClaim, ts=None,
                                       seqNo=None):
        if await self._wallet.shouldUpdateAccumulator(
                schemaId=ID(schemaId=schemaId), ts=ts,
                seqNo=seqNo):
            await self._wallet.updateAccumulator(schemaId=ID(schemaId=schemaId),
                                                 ts=ts,
                                                 seqNo=seqNo)

        oldV = c2.witness.V
        newAccum = await self._wallet.getAccumulator(
            ID(schemaId=schemaId))
        newV = newAccum.V
        tails = await self._wallet.getTails(ID(schemaId=schemaId))

        if c2.i not in newV:
            raise ValueError("Can not update Witness. I'm revoced.")

        if oldV != newV:
            vOldMinusNew = oldV - newV
            vNewMinusOld = newV - oldV
            omegaDenom = 1
            for j in vOldMinusNew:
                omegaDenom *= tails.gprime[newAccum.L + 1 - j + c2.i]
            omegaNum = 1
            newOmega = c2.witness.omega
            for j in vNewMinusOld:
                omegaNum *= tails.gprime[newAccum.L + 1 - j + c2.i]
                newOmega *= omegaNum / omegaDenom

            newWitness = c2.witness._replace(V=newV, omega=newOmega)
            c2 = c2._replace(witness=newWitness)

            await self._wallet.submitNonRevocClaim(schemaId=ID(schemaId=schemaId),
                                                   claim=c2)

        return c2

    async def initProof(self, schemaId,
                        c2: NonRevocationClaim) -> NonRevocInitProof:
        if not c2:
            return None

        c2 = await self.updateNonRevocationClaim(schemaId, c2)

        pkR = await self._wallet.getPublicKeyRevocation(ID(schemaId=schemaId))
        accum = await self._wallet.getAccumulator(ID(schemaId=schemaId))
        CList = []
        TauList = []

        cListParams = self._genCListParams(schemaId, c2)
        proofCList = self._createCListValues(schemaId, c2, cListParams, pkR)
        CList.extend(proofCList.asList())

        tauListParams = self._genTauListParams(schemaId)
        proofTauList = createTauListValues(pkR, accum, tauListParams,
                                           proofCList)
        TauList.extend(proofTauList.asList())

        return NonRevocInitProof(proofCList, proofTauList, cListParams,
                                 tauListParams)

    async def finalizeProof(self, schemaId, cH,
                            initProof: NonRevocInitProof) -> NonRevocProof:
        if not initProof:
            return None

        group = cmod.PairingGroup(
            PAIRING_GROUP)  # super singular curve, 1024 bits
        chNum_z = int_to_ZR(cH, group)
        XList = NonRevocProofXList.fromList(
            [x - chNum_z * y for x, y in zip(initProof.TauListParams.asList(),
                                             initProof.CListParams.asList())]
        )
        return NonRevocProof(XList, initProof.CList)

    def _genCListParams(self, schemaId,
                        c2: NonRevocationClaim) -> NonRevocProofXList:
        group = cmod.PairingGroup(
            PAIRING_GROUP)  # super singular curve, 1024 bits
        rho = group.random(cmod.ZR)
        r = group.random(cmod.ZR)
        rPrime = group.random(cmod.ZR)
        rPrimePrime = group.random(cmod.ZR)
        rPrimePrimePrime = group.random(cmod.ZR)
        o = group.random(cmod.ZR)
        oPrime = group.random(cmod.ZR)
        m = rho * c2.c
        mPrime = r * rPrimePrime
        t = o * c2.c
        tPrime = oPrime * rPrimePrime
        m2 = group.init(cmod.ZR, int(c2.m2))
        return NonRevocProofXList(rho=rho, r=r, rPrime=rPrime,
                                  rPrimePrime=rPrimePrime,
                                  rPrimePrimePrime=rPrimePrimePrime,
                                  o=o, oPrime=oPrime, m=m, mPrime=mPrime, t=t,
                                  tPrime=tPrime, m2=m2, s=c2.v, c=c2.c)

    def _createCListValues(self, schemaId, c2: NonRevocationClaim,
                           params: NonRevocProofXList,
                           pkR) -> NonRevocProofCList:
        E = (pkR.h ** params.rho) * (pkR.htilde ** params.o)
        D = (pkR.g ** params.r) * (pkR.htilde ** params.oPrime)
        A = c2.sigma * (pkR.htilde ** params.rho)
        G = c2.witness.gi * (pkR.htilde ** params.r)
        W = c2.witness.omega * (pkR.hhat ** params.rPrime)
        S = c2.witness.sigmai * (pkR.hhat ** params.rPrimePrime)
        U = c2.witness.ui * (pkR.hhat ** params.rPrimePrimePrime)
        return NonRevocProofCList(E, D, A, G, W, S, U)

    def _genTauListParams(self, schemaId) -> NonRevocProofXList:
        group = cmod.PairingGroup(
            PAIRING_GROUP)  # super singular curve, 1024 bits
        return NonRevocProofXList(group=group)

    async def testProof(self, schemaId, c2: NonRevocationClaim):
        pkR = await self._wallet.getPublicKeyRevocation(ID(schemaId=schemaId))
        accum = await self._wallet.getAccumulator(ID(schemaId=schemaId))
        accumPk = await self._wallet.getPublicKeyAccumulator(
            ID(schemaId=schemaId))

        cListParams = self._genCListParams(schemaId, c2)
        proofCList = self._createCListValues(schemaId, c2, cListParams, pkR)
        proofTauList = createTauListValues(pkR, accum, cListParams, proofCList)

        proofTauListCalc = createTauListExpectedValues(pkR, accum, accumPk,
                                                       proofCList)

        if proofTauListCalc.asList() != proofTauList.asList():
            raise ValueError("revocation proof is incorrect")

        return True



from abc import abstractmethod

from anoncreds.protocol.repo.public_repo import PublicRepo
from anoncreds.protocol.types import ID, \
    Claims, ClaimInitDataType, \
    PrimaryClaim, NonRevocationClaim, ClaimsPair, ClaimAttributeValues
from anoncreds.protocol.wallet.wallet import Wallet, WalletInMemory
from typing import Dict


class ProverWallet(Wallet):
    def __init__(self, schemaId, repo: PublicRepo):
        Wallet.__init__(self, schemaId, repo)

    # SUBMIT

    @abstractmethod
    async def submitClaimAttributes(
            self, schemaId: ID,
            claimAttributes: Dict[str, ClaimAttributeValues]):
        raise NotImplementedError

    @abstractmethod
    async def submitPrimaryClaim(self, schemaId: ID, claim: PrimaryClaim):
        raise NotImplementedError

    @abstractmethod
    async def submitNonRevocClaim(self, schemaId: ID,
                                  claim: NonRevocationClaim):
        raise NotImplementedError

    @abstractmethod
    async def submitMasterSecret(self, ms, schemaId: ID):
        raise NotImplementedError

    @abstractmethod
    async def submitPrimaryClaimInitData(self, schemaId: ID,
                                         claimInitData: ClaimInitDataType):
        raise NotImplementedError

    @abstractmethod
    async def submitNonRevocClaimInitData(self, schemaId: ID,
                                          claimInitData: ClaimInitDataType):
        raise NotImplementedError

    @abstractmethod
    async def submitContextAttr(self, schemaId: ID, m2):
        raise NotImplementedError

    # GET

    @abstractmethod
    async def getMasterSecret(self, schemaId: ID):
        raise NotImplementedError

    @abstractmethod
    async def getClaimAttributes(self, schemaId: ID):
        raise NotImplementedError

    @abstractmethod
    async def getAllClaimsAttributes(self) -> ClaimsPair:
        raise NotImplementedError

    @abstractmethod
    async def getClaimSignature(self, schemaId: ID) -> Claims:
        raise NotImplementedError

    @abstractmethod
    async def getAllClaimsSignatures(self):
        raise NotImplementedError

    @abstractmethod
    async def getPrimaryClaimInitData(self,
                                      schemaId: ID) -> ClaimInitDataType:
        raise NotImplementedError

    @abstractmethod
    async def getNonRevocClaimInitData(self,
                                       schemaId: ID) -> ClaimInitDataType:
        raise NotImplementedError

    @abstractmethod
    async def getContextAttr(self, schemaId: ID):
        raise NotImplementedError


class ProverWalletInMemory(ProverWallet, WalletInMemory):
    def __init__(self, schemaId, repo: PublicRepo):
        WalletInMemory.__init__(self, schemaId, repo)

        self._claims = {}

        # other dicts with key=schemaKey
        self._m1s = {}
        self._m2s = {}

        self._c1s = {}
        self._c2s = {}

        self._primaryInitData = {}
        self._nonRevocInitData = {}

    # SUBMIT

    async def submitClaimAttributes(
            self, schemaId: ID,
            claimAttributes: Dict[str, ClaimAttributeValues]):
        await self._cacheValueForId(self._claims, schemaId, claimAttributes)

    async def submitPrimaryClaim(self, schemaId: ID, claim: PrimaryClaim):
        await self._cacheValueForId(self._c1s, schemaId, claim)

    async def submitNonRevocClaim(self, schemaId: ID,
                                  claim: NonRevocationClaim):
        await self._cacheValueForId(self._c2s, schemaId, claim)

    async def submitMasterSecret(self, ms, schemaId: ID):
        await self._cacheValueForId(self._m1s, schemaId, ms)

    async def submitPrimaryClaimInitData(self, schemaId: ID,
                                         claimInitData: ClaimInitDataType):
        await self._cacheValueForId(self._primaryInitData, schemaId,
                                    claimInitData)

    async def submitNonRevocClaimInitData(self, schemaId: ID,
                                          claimInitData: ClaimInitDataType):
        await self._cacheValueForId(self._nonRevocInitData, schemaId,
                                    claimInitData)

    async def submitContextAttr(self, schemaId: ID, m2):
        await self._cacheValueForId(self._m2s, schemaId, m2)

    # GET

    async def getMasterSecret(self, schemaId: ID):
        return await self._getValueForId(self._m1s, schemaId)

    async def getClaimAttributes(self, schemaId: ID):
        return await self._getValueForId(self._claims, schemaId)

    async def getClaimSignature(self, schemaId: ID) -> Claims:
        c1 = await self._getValueForId(self._c1s, schemaId)
        c2 = None if not self._c2s else await self._getValueForId(self._c2s,
                                                                  schemaId)
        return Claims(c1, c2)

    async def getAllClaimsAttributes(self) -> ClaimsPair:
        res = ClaimsPair()
        for schemaKey in self._claims.keys():
            res[schemaKey] = await self.getClaimAttributes(ID(schemaKey))
        return res

    async def getAllClaimsSignatures(self):
        res = dict()
        for schemaKey in self._c1s.keys():
            res[schemaKey] = await self.getClaimSignature(ID(schemaKey))
        return res

    async def getPrimaryClaimInitData(self,
                                      schemaId: ID) -> ClaimInitDataType:
        return await self._getValueForId(self._primaryInitData, schemaId)

    async def getNonRevocClaimInitData(self,
                                       schemaId: ID) -> ClaimInitDataType:
        return await self._getValueForId(self._nonRevocInitData, schemaId)

    async def getContextAttr(self, schemaId: ID):
        return await self._getValueForId(self._m2s, schemaId)



from typing import Dict

from anoncreds.protocol.globals import LARGE_MASTER_SECRET
from anoncreds.protocol.primary.primary_claim_issuer import PrimaryClaimIssuer
from anoncreds.protocol.repo.attributes_repo import AttributeRepo
from anoncreds.protocol.revocation.accumulators.non_revocation_claim_issuer import \
    NonRevocationClaimIssuer
from anoncreds.protocol.types import PrimaryClaim, NonRevocationClaim, \
    Schema, ID, Claims, ClaimRequest, Attribs, PublicKey, \
    RevocationPublicKey, AccumulatorPublicKey, ClaimAttributeValues
from anoncreds.protocol.utils import strToInt, get_hash_as_int
from anoncreds.protocol.wallet.issuer_wallet import IssuerWallet
from config.config import cmod


class Issuer:
    def __init__(self, wallet: IssuerWallet, attrRepo: AttributeRepo):
        self.wallet = wallet
        self._attrRepo = attrRepo
        self._primaryIssuer = PrimaryClaimIssuer(wallet)
        self._nonRevocationIssuer = NonRevocationClaimIssuer(wallet)

    #
    # PUBLIC
    #

    @property
    def issuerId(self):
        return self.wallet.walletId

    def isSchemaExists(self, schemaKey):
        return self.wallet._schemasByKey.get(schemaKey)

    async def genSchema(self, name, version, attrNames) -> Schema:
        """
        Generates and submits Schema.

        :param name: schema name
        :param version: schema version
        :param attrNames: a list of attributes the schema contains
        :return: submitted Schema
        """
        schema = Schema(name, version, attrNames, self.issuerId)
        return await self.wallet.submitSchema(schema)

    async def genKeys(self, schemaId: ID, p_prime=None, q_prime=None) -> (
            PublicKey, RevocationPublicKey):
        """
        Generates and submits keys (both public and secret, primary and
        non-revocation).

        :param schemaId: The schema ID (reference to claim
        definition schema)
        :param p_prime: optional p_prime parameter
        :param q_prime: optional q_prime parameter
        :return: Submitted Public keys (both primary and non-revocation)
        """
        pk, sk = await self._primaryIssuer.genKeys(schemaId, p_prime, q_prime)
        pkR, skR = await self._nonRevocationIssuer.genRevocationKeys()
        pk = await self.wallet.submitPublicKeys(schemaId=schemaId, pk=pk,
                                                pkR=pkR)
        pkR = await self.wallet.submitSecretKeys(schemaId=schemaId, sk=sk,
                                                 skR=skR)
        return pk, pkR

    async def issueAccumulator(self, schemaId: ID, iA,
                               L) -> AccumulatorPublicKey:
        """
        Issues and submits an accumulator used for non-revocation proof.

        :param schemaId: The schema ID (reference to claim
        definition schema)
        :param iA: accumulator ID
        :param L: maximum number of claims within accumulator.
        :return: Submitted accumulator public key
        """
        accum, tails, accPK, accSK = await self._nonRevocationIssuer.issueAccumulator(
            schemaId, iA, L)
        accPK = await self.wallet.submitAccumPublic(schemaId=schemaId,
                                                    accumPK=accPK,
                                                    accum=accum, tails=tails)
        await self.wallet.submitAccumSecret(schemaId=schemaId,
                                            accumSK=accSK)
        return accPK

    async def revoke(self, schemaId: ID, i):
        """
        Performs revocation of a Claim.

        :param schemaId: The schema ID (reference to claim
        definition schema)
        :param i: claim's sequence number within accumulator
        """
        acc, ts = await self._nonRevocationIssuer.revoke(schemaId, i)
        await self.wallet.submitAccumUpdate(schemaId=schemaId, accum=acc,
                                            timestampMs=ts)

    async def issueClaim(self, schemaId: ID, claimRequest: ClaimRequest,
                         iA=None,
                         i=None) -> (Claims, Dict[str, ClaimAttributeValues]):
        """
        Issue a claim for the given user and schema.

        :param schemaId: The schema ID (reference to claim
        definition schema)
        :param claimRequest: A claim request containing prover ID and
        prover-generated values
        :param iA: accumulator ID
        :param i: claim's sequence number within accumulator
        :return: The claim (both primary and non-revocation)
        """

        schemaKey = (await self.wallet.getSchema(schemaId)).getKey()
        attributes = self._attrRepo.getAttributes(schemaKey,
                                                  claimRequest.userId)

        # TODO re-enable when revocation registry is implemented
        # iA = iA if iA else (await self.wallet.getAccumulator(schemaId)).iA

        # TODO this has un-obvious side-effects
        await self._genContxt(schemaId, iA, claimRequest.userId)

        (c1, claim) = await self._issuePrimaryClaim(schemaId, attributes,
                                                    claimRequest.U)
        # TODO re-enable when revocation registry is fully implemented
        c2 = await self._issueNonRevocationClaim(schemaId, claimRequest.Ur,
                                                 iA,
                                                 i) if claimRequest.Ur else None

        signature = Claims(primaryClaim=c1, nonRevocClaim=c2)

        return (signature, claim)

    async def issueClaims(self, allClaimRequest: Dict[ID, ClaimRequest]) -> \
            Dict[ID, Claims]:
        """
        Issue claims for the given users and schemas.

        :param allClaimRequest: a map of schema ID to a claim
        request containing prover ID and prover-generated values
        :return: The claims (both primary and non-revocation)
        """
        res = {}
        for schemaId, claimReq in allClaimRequest.items():
            res[schemaId] = await self.issueClaim(schemaId, claimReq)
        return res

    #
    # PRIVATE
    #

    async def _genContxt(self, schemaId: ID, iA, userId):
        iA = strToInt(str(iA))
        userId = strToInt(str(userId))
        S = iA | userId
        H = get_hash_as_int(S)
        m2 = cmod.integer(H % (2 ** LARGE_MASTER_SECRET))
        await self.wallet.submitContextAttr(schemaId, m2)
        return m2

    async def _issuePrimaryClaim(self, schemaId: ID, attributes: Attribs,
                                 U) -> (PrimaryClaim, Dict[str, ClaimAttributeValues]):
        return await self._primaryIssuer.issuePrimaryClaim(schemaId,
                                                           attributes, U)

    async def _issueNonRevocationClaim(self, schemaId: ID, Ur, iA=None,
                                       i=None) -> NonRevocationClaim:
        claim, accum, ts = await self._nonRevocationIssuer.issueNonRevocationClaim(
            schemaId, Ur, iA, i)
        await self.wallet.submitAccumUpdate(schemaId=schemaId, accum=accum,
                                            timestampMs=ts)
        return claim

    def __repr__(self):
        return str(self.__dict__)



from collections import OrderedDict

import pytest

from anoncreds.protocol.globals import PAIRING_GROUP
from anoncreds.protocol.utils import toDictWithStrValues, \
    deserializeFromStr, serializeToStr, fromDictWithStrValues, get_hash_as_int, intToArrayBytes, bytesToInt
from anoncreds.test.conftest import primes
from config.config import cmod


def testStrSerializeToFromStr():
    value = 'aaa'
    assert value == deserializeFromStr(serializeToStr(value))


def testIntSerializeToFromStr():
    value = 111
    assert value == deserializeFromStr(serializeToStr(value))


def testCryptoIntSerializeToFromStr():
    value = cmod.integer(44444444444444444)
    assert value == deserializeFromStr(serializeToStr(value))


def testCryptoIntModSerializeToFromStr():
    value = cmod.integer(44444444444444444) % 33
    assert value == deserializeFromStr(serializeToStr(value))


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
def testGroupElementSerializeToFromStr():
    value = cmod.PairingGroup(PAIRING_GROUP).random(cmod.G1)
    assert value == deserializeFromStr(serializeToStr(value))


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
def testGroupElementZRIdentitySerializeToFromStr():
    elem = cmod.PairingGroup(PAIRING_GROUP).init(cmod.ZR, 555)
    identity = elem / elem
    assert identity == deserializeFromStr(serializeToStr(identity))


def testGroupElementG1IdentitySerializeToFromStr():
    elem = cmod.PairingGroup(PAIRING_GROUP).random(cmod.G1)
    identity = cmod.PairingGroup(PAIRING_GROUP).init(cmod.G1, elem / elem)
    assert identity == deserializeFromStr(serializeToStr(identity))


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
def testToFromDictWithStrValues():
    group = cmod.PairingGroup(PAIRING_GROUP)
    dictionary = OrderedDict((
        ('43', '43'),
        ('3', 3),
        ('5', cmod.integer(111) % 11),
        ('10', group.random(cmod.G1))
    ))
    assert dictionary == fromDictWithStrValues(toDictWithStrValues(dictionary))


def testToFromDictWithStrValuesInteKeys():
    dictionary = OrderedDict((
        (11, '43'),
        (12, 3)
    ))
    assert dictionary == fromDictWithStrValues(toDictWithStrValues(dictionary))


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
def testToFromDictWithStrValuesLists():
    group = cmod.PairingGroup(PAIRING_GROUP)
    dictionary = OrderedDict((
        ('47', []),
        ('7',
         [cmod.integer(111) % 11, cmod.integer(222), cmod.integer(333) % 45]),
        ('6', [group.init(cmod.ZR, 555), group.random(cmod.G1),
               group.random(cmod.G1)])
    ))
    assert dictionary == fromDictWithStrValues(toDictWithStrValues(dictionary))


def testToFromDictWithStrValuesSets():
    dictionary = OrderedDict((
        ('44', {'aaa', 'bbb'}),
        ('4', {111, 2222}),
        ('1', {}),
        ('3', 3),
    ))
    assert dictionary == fromDictWithStrValues(toDictWithStrValues(dictionary))


def testToFromDictWithStrValuesSubDicts():
    cmod.PairingGroup(PAIRING_GROUP)
    dictionary = OrderedDict((
        ('4', {'aaa', 'bbb'}),
        ('2', OrderedDict((
            ('33',
             OrderedDict((('45', 45), ('11', 11)))
             ),
            ('23',
             OrderedDict((('47', 47), ('34', 34)))
             )
        ))),
        ('3', 3)
    ))
    assert dictionary == fromDictWithStrValues(toDictWithStrValues(dictionary))


@pytest.mark.skipif('sys.platform == "win32"', reason='SOV-86')
def testToFromDictWithStrValuesMixed():
    group = cmod.PairingGroup(PAIRING_GROUP)
    dictionary = OrderedDict((
        ('4', {'aaa', 'bbb'}),
        ('2', OrderedDict((
            ('33',
             OrderedDict((('45', 45), ('11', 11)))
             ),
            ('23',
             OrderedDict((('47', 47), ('34', 34)))
             )
        ))),
        ('1', {}),
        ('3', 3),
        ('5', cmod.integer(111) % 11),
        ('7',
         [cmod.integer(111) % 11, cmod.integer(222), cmod.integer(333) % 45]),
        ('6', [group.init(cmod.ZR, 555), group.random(cmod.G1),
               group.random(cmod.G1)]),
        ('10', group.random(cmod.G1))
    ))
    assert dictionary == fromDictWithStrValues(toDictWithStrValues(dictionary))


def testToFromDictWithStrValuesDictInList():
    dictionary = OrderedDict((
        ('2', [
            OrderedDict((
                ('33',
                 OrderedDict((('45', 45), ('11', 11)))
                 ),
                ('23',
                 OrderedDict((('47', 47), ('34', 34)))
                 )
            )),
            OrderedDict((
                ('63',
                 OrderedDict((('65', 45), ('61', 11)))
                 ),
                ('63',
                 OrderedDict((('67', 47), ('64', 34)))
                 )
            ))
        ]
        ),
        ('3', 3)
    ))
    assert dictionary == fromDictWithStrValues(toDictWithStrValues(dictionary))


def testGetHashInt():
    input = [0xb1134a647eb818069c089e7694f63e6d,
             0x57fbf9dc8c8e6acde33de98c6d747b28c,
             0x77fbf9dc8c8e6acde33de98c6d747b28c]

    _checkHashEqual(input)


def testGetHashInteger():
    P_PRIME1, Q_PRIME1 = primes.get("prime1")
    P_PRIME2, Q_PRIME2 = primes.get("prime2")
    input = [P_PRIME1, Q_PRIME1, P_PRIME2, Q_PRIME2]

    _checkHashEqual(input)


def testGetHashGroup():
    group = cmod.PairingGroup(PAIRING_GROUP)
    input = [group.random(cmod.G1),
             group.random(cmod.G1),
             group.random(cmod.G1)]

    _checkHashEqual(input)


def testGetHashMixed():
    group = cmod.PairingGroup(PAIRING_GROUP)
    P_PRIME1, Q_PRIME1 = primes.get("prime1")
    input = [P_PRIME1, Q_PRIME1,
             group.random(cmod.G1), group.random(cmod.G1),
             0xb1134a647eb818069c089e7694f63e6d,
             0x57fbf9dc8c8e6acde33de98c6d747b28c]

    _checkHashEqual(input)


def _checkHashEqual(input):
    h1 = get_hash_as_int(*input)
    h2 = get_hash_as_int(*reversed(input))
    assert h1 == h2


def testIntToArrayBytes():
    val = cmod.integer(1606507817390189252221968804450207070282033)
    res = [18, 113, 26, 39, 35, 240, 231, 239, 92,
           226, 84, 46, 230, 174, 230, 41, 225, 49]
    assert res == intToArrayBytes(val)


def testBytesToInt():
    val = [18, 113, 26, 39, 35, 240, 231, 239, 92,
           226, 84, 46, 230, 174, 230, 41, 225, 49]
    res = 1606507817390189252221968804450207070282033
    assert res == bytesToInt(val)


def testIntToArrayBytesAndBack():
    val = cmod.integer(1606507817390189252221968804450207070282033)
    assert val == bytesToInt(intToArrayBytes(val))



from typing import Sequence, Dict

from anoncreds.protocol.globals import LARGE_VPRIME, LARGE_MVECT, LARGE_E_START, \
    LARGE_ETILDE, \
    LARGE_VTILDE, LARGE_UTILDE, LARGE_RTILDE, LARGE_ALPHATILDE, ITERATIONS, \
    DELTA
from anoncreds.protocol.primary.primary_proof_common import calcTge, calcTeq
from anoncreds.protocol.types import PrimaryClaim, Predicate, PrimaryInitProof, \
    PrimaryEqualInitProof, PrimaryPrecicateGEInitProof, PrimaryProof, \
    PrimaryEqualProof, PrimaryPredicateGEProof, \
    ID, ClaimInitDataType, ClaimAttributeValues
from anoncreds.protocol.utils import splitRevealedAttrs, fourSquares
from anoncreds.protocol.wallet.prover_wallet import ProverWallet
from config.config import cmod


class PrimaryClaimInitializer:
    def __init__(self, wallet: ProverWallet):
        self._wallet = wallet

    async def genClaimInitData(self, schemaId: ID) -> ClaimInitDataType:
        pk = await self._wallet.getPublicKey(schemaId)
        ms = await self._wallet.getMasterSecret(schemaId)
        vprime = cmod.randomBits(LARGE_VPRIME)
        N = pk.N
        Rms = pk.Rms
        S = pk.S
        U = (S ** vprime) * (Rms ** ms) % N

        return ClaimInitDataType(U=U, vPrime=vprime)

    async def preparePrimaryClaim(self, schemaId: ID, claim: PrimaryClaim):
        claimInitDat = await self._wallet.getPrimaryClaimInitData(schemaId)
        newV = claim.v + claimInitDat.vPrime
        claim = claim._replace(v=newV)
        return claim


class PrimaryProofBuilder:
    def __init__(self, wallet: ProverWallet):
        self._wallet = wallet

    async def initProof(self, schemaId, c1: PrimaryClaim,
                        revealedAttrs: Sequence[str],
                        predicates: Sequence[Predicate],
                        m1Tilde, m2Tilde, claimAttributes: Dict[str, ClaimAttributeValues]) -> PrimaryInitProof:
        if not c1:
            return None

        eqProof = await self._initEqProof(schemaId, c1, revealedAttrs,
                                          m1Tilde, m2Tilde, claimAttributes)
        geProofs = []
        for predicate in predicates:
            geProof = await self._initGeProof(schemaId, eqProof, c1,
                                              predicate, claimAttributes)
            geProofs.append(geProof)
        return PrimaryInitProof(eqProof, geProofs)

    async def finalizeProof(self, schemaId, cH,
                            initProof: PrimaryInitProof) -> PrimaryProof:
        if not initProof:
            return None

        cH = cmod.integer(cH)
        eqProof = await self._finalizeEqProof(schemaId, cH,
                                              initProof.eqProof)
        geProofs = []
        for initGeProof in initProof.geProofs:
            geProof = await self._finalizeGeProof(schemaId, cH, initGeProof,
                                                  eqProof)
            geProofs.append(geProof)
        return PrimaryProof(eqProof, geProofs)

    async def _initEqProof(self, schemaId, c1: PrimaryClaim,
                           revealedAttrs: Sequence[str], m1Tilde, m2Tilde, claimAttributes: Dict[str, ClaimAttributeValues]) \
            -> PrimaryEqualInitProof:
        m2Tilde = m2Tilde if m2Tilde else cmod.integer(
            cmod.randomBits(LARGE_MVECT))
        revealedAttrs, unrevealedAttrs = splitRevealedAttrs(
            claimAttributes, [a.name for a in revealedAttrs])
        mtilde = self._getMTilde(unrevealedAttrs)

        Ra = cmod.integer(cmod.randomBits(LARGE_VPRIME))
        pk = await self._wallet.getPublicKey(ID(schemaId=schemaId))

        A, e, v = c1.A, c1.e, c1.v
        Aprime = A * (pk.S ** Ra) % pk.N
        vprime = (v - e * Ra)
        eprime = e - (2 ** LARGE_E_START)

        etilde = cmod.integer(cmod.randomBits(LARGE_ETILDE))
        vtilde = cmod.integer(cmod.randomBits(LARGE_VTILDE))

        Rur = 1 % pk.N
        for k, value in unrevealedAttrs.items():
            if k in claimAttributes:
                Rur = Rur * (pk.R[k] ** mtilde[k])
        Rur *= pk.Rms ** m1Tilde
        Rur *= pk.Rctxt ** m2Tilde

        # T = ((Aprime ** etilde) * Rur * (pk.S ** vtilde)) % pk.N
        T = calcTeq(pk, Aprime, etilde, vtilde, mtilde, m1Tilde, m2Tilde,
                    unrevealedAttrs.keys())

        return PrimaryEqualInitProof(c1, Aprime, T, etilde, eprime, vtilde,
                                     vprime, mtilde, m1Tilde, m2Tilde,
                                     unrevealedAttrs.keys(), revealedAttrs)

    async def _initGeProof(self, schemaId, eqProof: PrimaryEqualInitProof,
                           c1: PrimaryClaim, predicate: Predicate, claimAttributes: Dict[str, ClaimAttributeValues]) \
            -> PrimaryPrecicateGEInitProof:
        # gen U for Delta
        pk = await self._wallet.getPublicKey(ID(schemaId=schemaId))
        k, value = predicate.attrName, predicate.value
        delta = claimAttributes[k].encoded - value
        if delta < 0:
            raise ValueError("Predicate is not satisfied")

        u = fourSquares(delta)

        # prepare C list
        r = {}
        T = {}
        CList = []
        for i in range(0, ITERATIONS):
            r[str(i)] = cmod.integer(cmod.randomBits(LARGE_VPRIME))
            T[str(i)] = (pk.Z ** u[str(i)]) * (pk.S ** r[str(i)]) % pk.N
            CList.append(T[str(i)])
        r[DELTA] = cmod.integer(cmod.randomBits(LARGE_VPRIME))
        T[DELTA] = (pk.Z ** delta) * (pk.S ** r[DELTA]) % pk.N
        CList.append(T[DELTA])

        # prepare Tau List
        utilde = {}
        rtilde = {}
        for i in range(0, ITERATIONS):
            utilde[str(i)] = cmod.integer(cmod.randomBits(LARGE_UTILDE))
            rtilde[str(i)] = cmod.integer(cmod.randomBits(LARGE_RTILDE))
        rtilde[DELTA] = cmod.integer(cmod.randomBits(LARGE_RTILDE))
        alphatilde = cmod.integer(cmod.randomBits(LARGE_ALPHATILDE))

        TauList = calcTge(pk, utilde, rtilde, eqProof.mTilde[k], alphatilde, T)
        return PrimaryPrecicateGEInitProof(CList, TauList, u, utilde, r, rtilde,
                                           alphatilde, predicate, T)

    async def _finalizeEqProof(self, schemaId, cH,
                               initProof: PrimaryEqualInitProof) -> PrimaryEqualProof:
        e = initProof.eTilde + (cH * initProof.ePrime)
        v = initProof.vTilde + (cH * initProof.vPrime)

        m = {}

        claimAttributes = await self._wallet.getClaimAttributes(ID(schemaId=schemaId))

        for k in initProof.unrevealedAttrs:
            m[str(k)] = initProof.mTilde[str(k)] + (
                cH * claimAttributes[str(k)].encoded)
        ms = await self._wallet.getMasterSecret(ID(schemaId=schemaId))
        m1 = initProof.m1Tilde + (cH * ms)
        m2 = initProof.m2Tilde + (cH * initProof.c1.m2)

        return PrimaryEqualProof(e, v, m, m1, m2, initProof.Aprime,
                                 initProof.revealedAttrs)

    async def _finalizeGeProof(self, schemaId, cH,
                               initProof: PrimaryPrecicateGEInitProof,
                               eqProof: PrimaryEqualProof) \
            -> PrimaryPredicateGEProof:
        u = {}
        r = {}
        urproduct = 0
        for i in range(0, ITERATIONS):
            u[str(i)] = initProof.uTilde[str(i)] + cH * initProof.u[str(i)]
            r[str(i)] = initProof.rTilde[str(i)] + cH * initProof.r[str(i)]
            urproduct += initProof.u[str(i)] * initProof.r[str(i)]
            r[DELTA] = initProof.rTilde[DELTA] + cH * initProof.r[DELTA]

        alpha = initProof.alphaTilde + cH * (initProof.r[DELTA] - urproduct)

        k = initProof.predicate.attrName
        return PrimaryPredicateGEProof(u, r, alpha, eqProof.m[str(k)],
                                       initProof.T, initProof.predicate)

    def _getMTilde(self, unrevealedAttrs):
        mtilde = {}
        for key, value in unrevealedAttrs.items():
            mtilde[key] = cmod.integer(cmod.randomBits(LARGE_MVECT))
        return mtilde
